from __future__ import annotations

import abc
from datetime import datetime
import json
import logging
from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import pandas as pd
from trulens.core.schema import app as app_schema
from trulens.core.schema import dataset as dataset_schema
from trulens.core.schema import event as event_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import groundtruth as groundtruth_schema
from trulens.core.schema import record as record_schema
from trulens.core.schema import types as types_schema
from trulens.core.schema.event import Event
from trulens.core.utils import json as json_utils
from trulens.core.utils import serial as serial_utils
from trulens.core.utils import text as text_utils
from trulens.otel.semconv.trace import BASE_SCOPE
from trulens.otel.semconv.trace import ResourceAttributes
from trulens.otel.semconv.trace import SpanAttributes

NoneType = type(None)

logger = logging.getLogger(__name__)

MULTI_CALL_NAME_DELIMITER = ":::"

DEFAULT_DATABASE_PREFIX: str = "trulens_"
"""Default prefix for table names for trulens to use.

This includes alembic's version table.
"""

DEFAULT_DATABASE_FILE: str = "default.sqlite"
"""Filename for default sqlite database.

The sqlalchemy url for this default local sqlite database is `sqlite:///default.sqlite`.
"""

DEFAULT_DATABASE_REDACT_KEYS: bool = False
"""Default value for option to redact secrets before writing out data to database."""


class BaseAppsExtractor:
    """Utilities for creating dataframes from orm instances."""

    app_cols = ["app_name", "app_version", "app_id", "app_json", "type"]
    rec_cols = [
        "record_id",
        "input",
        "output",
        "tags",
        "record_json",
        "cost_json",
        "perf_json",
        "ts",
    ]
    extra_cols = [
        "latency",
        "total_tokens",
        "total_cost",
        "eval_cost",
        "eval_cost_snowflake",
        "num_events",
    ]
    all_cols = app_cols + rec_cols + extra_cols


class DB(serial_utils.SerialModel, abc.ABC, text_utils.WithIdentString):
    """Abstract definition of databases used by trulens.

    [SQLAlchemyDB][trulens.core.database.sqlalchemy.SQLAlchemyDB] is the main
    and default implementation of this interface.
    """

    redact_keys: bool = DEFAULT_DATABASE_REDACT_KEYS
    """Redact secrets before writing out data."""

    table_prefix: str = DEFAULT_DATABASE_PREFIX
    """Prefix for table names for trulens to use.

    May be useful in some databases where trulens is not the only app.
    """

    def __str__(self) -> str:
        return f"{self.__class__.__name__}"

    # WithIdentString requirement:
    def _ident_str(self) -> str:
        return f"{self.__class__.__name__}"

    def _json_str_of_obj(self, obj: Any) -> str:
        return json_utils.json_str_of_obj(obj, redact_keys=self.redact_keys)

    @abc.abstractmethod
    def reset_database(self):
        """Delete all data."""

        raise NotImplementedError()

    @abc.abstractmethod
    def migrate_database(self, prior_prefix: Optional[str] = None):
        """Migrate the stored data to the current configuration of the database.

        Args:
            prior_prefix: If given, the database is assumed to have been
                reconfigured from a database with the given prefix. If not
                given, it may be guessed if there is only one table in the
                database with the suffix `alembic_version`.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def check_db_revision(self):
        """Check that the database is up to date with the current trulens
        version.

        Raises:
            ValueError: If the database is not up to date.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_db_revision(self) -> Optional[str]:
        """Get the current revision of the database.

        Returns:
            The current revision of the database.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def insert_record(
        self,
        record: record_schema.Record,
    ) -> types_schema.RecordID:
        """
        Upsert a `record` into the database.

        Args:
            record: The record to insert or update.

        Returns:
            The id of the given record.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def batch_insert_record(
        self, records: List[record_schema.Record]
    ) -> List[types_schema.RecordID]:
        """
        Upsert a batch of records into the database.

        Args:
            records: The records to insert or update.

        Returns:
            The ids of the given records.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def insert_app(self, app: app_schema.AppDefinition) -> types_schema.AppID:
        """
        Upsert an `app` into the database.

        Args:
            app: The app to insert or update. Note that only the
                [AppDefinition][trulens.core.schema.app.AppDefinition] parts are serialized
                hence the type hint.

        Returns:
            The id of the given app.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def delete_app(self, app_id: types_schema.AppID) -> None:
        """
        Delete an `app` from the database.

        Args:
            app_id: The id of the app to delete.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def insert_feedback_definition(
        self, feedback_definition: feedback_schema.FeedbackDefinition
    ) -> types_schema.FeedbackDefinitionID:
        """
        Upsert a `feedback_definition` into the database.

        Args:
            feedback_definition: The feedback definition to insert or update.
                Note that only the
                [FeedbackDefinition][trulens.core.schema.feedback.FeedbackDefinition]
                parts are serialized hence the type hint.

        Returns:
            The id of the given feedback definition.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def get_feedback_defs(
        self,
        feedback_definition_id: Optional[
            types_schema.FeedbackDefinitionID
        ] = None,
    ) -> pd.DataFrame:
        """Retrieve feedback definitions from the database.

        Args:
            feedback_definition_id: if provided, only the
                feedback definition with the given id is returned. Otherwise,
                all feedback definitions are returned.

        Returns:
            A dataframe with the feedback definitions.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def insert_feedback(
        self,
        feedback_result: feedback_schema.FeedbackResult,
    ) -> types_schema.FeedbackResultID:
        """Upsert a `feedback_result` into the the database.

        Args:
            feedback_result: The feedback result to insert or update.

        Returns:
            The id of the given feedback result.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def batch_insert_feedback(
        self, feedback_results: List[feedback_schema.FeedbackResult]
    ) -> List[types_schema.FeedbackResultID]:
        """Upsert a batch of feedback results into the database.

        Args:
            feedback_results: The feedback results to insert or update.

        Returns:
            The ids of the given feedback results.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def get_feedback(
        self,
        record_id: Optional[types_schema.RecordID] = None,
        feedback_result_id: Optional[types_schema.FeedbackResultID] = None,
        feedback_definition_id: Optional[
            types_schema.FeedbackDefinitionID
        ] = None,
        status: Optional[
            Union[
                feedback_schema.FeedbackResultStatus,
                Sequence[feedback_schema.FeedbackResultStatus],
            ]
        ] = None,
        last_ts_before: Optional[datetime] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        shuffle: Optional[bool] = None,
        run_location: Optional[feedback_schema.FeedbackRunLocation] = None,
    ) -> pd.DataFrame:
        """Get feedback results matching a set of optional criteria:

        Args:
            record_id: Get only the feedback for the given record id.

            feedback_result_id: Get only the feedback for the given feedback
                result id.

            feedback_definition_id: Get only the feedback for the given feedback
                definition id.

            status: Get only the feedback with the given status. If a sequence
                of statuses is given, all feedback with any of the given
                statuses are returned.

            last_ts_before: get only results with `last_ts` before the
                given datetime.

            offset: index of the first row to return.

            limit: limit the number of rows returned.

            shuffle: shuffle the rows before returning them.

            run_location: Only get feedback functions with this run_location.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def get_feedback_count_by_status(
        self,
        record_id: Optional[types_schema.RecordID] = None,
        feedback_result_id: Optional[types_schema.FeedbackResultID] = None,
        feedback_definition_id: Optional[
            types_schema.FeedbackDefinitionID
        ] = None,
        status: Optional[
            Union[
                feedback_schema.FeedbackResultStatus,
                Sequence[feedback_schema.FeedbackResultStatus],
            ]
        ] = None,
        last_ts_before: Optional[datetime] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
        shuffle: bool = False,
        run_location: Optional[feedback_schema.FeedbackRunLocation] = None,
    ) -> Dict[feedback_schema.FeedbackResultStatus, int]:
        """Get count of feedback results matching a set of optional criteria grouped by
        their status.

        See [get_feedback][trulens.core.database.base.DB.get_feedback] for the meaning of
        the the arguments.

        Returns:
            A mapping of status to the count of feedback results of that status
                that match the given filters.
        """

        raise NotImplementedError()

    @abc.abstractmethod
    def get_app(
        self, app_id: types_schema.AppID
    ) -> Optional[serial_utils.JSONized]:
        """Get the app with the given id from the database.

        Returns:
            The jsonized version of the app with the given id. Deserialization
                can be done with
                [App.model_validate][trulens.core.app.App.model_validate].

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_apps(
        self, app_name: Optional[types_schema.AppName] = None
    ) -> Iterable[serial_utils.JSONized[app_schema.AppDefinition]]:
        """Get all apps."""

        raise NotImplementedError()

    def update_app_metadata(
        self, app_id: types_schema.AppID, metadata: Dict[str, Any]
    ) -> Optional[app_schema.AppDefinition]:
        """Update the metadata of an app."""

    @abc.abstractmethod
    def get_records_and_feedback(
        self,
        app_ids: Optional[List[types_schema.AppID]] = None,
        app_name: Optional[types_schema.AppName] = None,
        app_version: Optional[types_schema.AppVersion] = None,
        app_versions: Optional[List[types_schema.AppVersion]] = None,
        run_name: Optional[types_schema.RunName] = None,
        record_ids: Optional[List[types_schema.RecordID]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> Tuple[pd.DataFrame, Sequence[str]]:
        """Get records, their feedback results, and feedback names.

        Args:
            app_ids:
                A list of app ids to filter records by. If empty or not given,
                all apps' records will be returned.
            app_name:
                A name of the app to filter records by. If given, only records
                for this app will be returned.
            app_version:
                A version of the app to filter records by. If given, only
                records for this app version will be returned.
            run_name:
                A run name to filter records by. If given, only records for
                this run will be returned.
            app_versions:
                A list of app versions to filter records by. If given, only
                records for these app versions will be returned.
            record_ids: An optional list of record ids to filter records by.
            offset: Record row offset.
            limit: Limit on the number of records to return.

        Returns:
            Tuple of:
            - DataFrame of records with their feedback results.
            - List of feedback names that are columns in the DataFrame.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def insert_ground_truth(
        self, ground_truth: groundtruth_schema.GroundTruth
    ) -> types_schema.GroundTruthID:
        """Insert a ground truth entry into the database. The ground truth id is generated
        based on the ground truth content, so re-inserting is idempotent.

        Args:
            ground_truth: The ground truth entry to insert.

        Returns:
            The id of the given ground truth entry.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def batch_insert_ground_truth(
        self, ground_truths: List[groundtruth_schema.GroundTruth]
    ) -> List[types_schema.GroundTruthID]:
        """Insert a batch of ground truth entries into the database.

        Args:
            ground_truths: The ground truth entries to insert.

        Returns:
            The ids of the given ground truth entries.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_ground_truth(
        self,
        ground_truth_id: Optional[types_schema.GroundTruthID] = None,
    ) -> Optional[serial_utils.JSONized]:
        """Get the ground truth with the given id from the database."""

        raise NotImplementedError()

    @abc.abstractmethod
    def get_ground_truths_by_dataset(self, dataset_name: str) -> pd.DataFrame:
        """Get all ground truths from the database from a particular dataset's name.

        Returns:
            A dataframe with the ground truths.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_virtual_ground_truth(
        self,
        user_table_name: str,
        user_schema_mapping: Dict[str, str],
        user_schema_name: Optional[str] = None,
    ) -> pd.DataFrame:
        """Get all virtual ground truths from the database from a particular user table's name.

        Returns:
            A dataframe with the virtual ground truths.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def insert_dataset(
        self, dataset: dataset_schema.Dataset
    ) -> types_schema.DatasetID:
        """Insert a dataset into the database. The dataset id is generated based on the
        dataset content, so re-inserting is idempotent.

        Args:
            dataset: The dataset to insert.

        Returns:
            The id of the given dataset.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_datasets(self) -> pd.DataFrame:
        """Get all datasets from the database.

        Returns:
            A dataframe with the datasets.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def insert_event(self, event: event_schema.Event) -> types_schema.EventID:
        """Insert an event into the database.

        Args:
            event: The event to insert.

        Returns:
            The id of the given event.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def get_events(
        self,
        app_name: Optional[str],
        app_version: Optional[str],
        record_ids: Optional[List[str]],
        start_time: Optional[datetime],
    ) -> pd.DataFrame:
        """
        Get events from the database.

        Args:
            app_name: The app name to filter events by.
            app_version: The app version to filter events by.
            record_ids: The record ids to filter events by.
            start_time: The minimum time to consider events from.

        Returns:
            A pandas DataFrame of all relevant events.
        """
        raise NotImplementedError()

    def _get_records_and_feedback_otel_from_events(
        self,
        events: List[Event],
        app_ids: Optional[List[str]] = None,
        app_name: Optional[types_schema.AppName] = None,
    ) -> Tuple[pd.DataFrame, Sequence[str]]:
        if not events:
            # Return empty dataframe with expected columns
            logger.warning(
                f"No events found for app_name: {app_name}, app_ids: {app_ids}"
            )
            return pd.DataFrame(columns=BaseAppsExtractor.all_cols), []

        # Group events by record_id
        record_events = {}
        for event in events:
            record_attributes = self._get_event_record_attributes_otel(event)
            resource_attributes = self._get_event_resource_attributes_otel(
                event
            )
            record_id = record_attributes.get(SpanAttributes.RECORD_ID)
            if not record_id:
                continue
            app_name, app_version, app_id, _ = self.extract_app_and_run_info(
                record_attributes, resource_attributes
            )
            if app_ids and app_id not in app_ids:
                # TODO(otel):
                # This may screw up the pagination and can be slow due to it
                # looking at possibly a lot more events if there are many
                # that don't have app ids.
                # In the future we should either:
                # 1. Remove app ids if we're going to assume they're some
                #    complex function of app_name and app_version that's
                #    hard to replicate for non-TruLens users that want to
                #    still use our evaluation/feedback stuff.
                # 2. Have the app ids be from some source of truth like the
                #    app table but this doesn't work as easily for the
                #    Snowflake side.
                logger.info(f"Computed {app_id} not in {app_ids}!")
                continue

            if record_id not in record_events:
                record_events[record_id] = {
                    "events": [],
                    "app_name": app_name,
                    "app_version": app_version,
                    "app_id": app_id,
                    "input": "",  # Initialize to empty string, filled below
                    "output": "",  # Initialize to empty string, filled below
                    "tags": "",  # Not present in OTEL, use empty string
                    "ts": pd.NaT,  # Initialize to empty value, filled below
                    "latency": 0.0,  # Initialize to 0.0, filled below
                    "total_tokens": 0,  # Initialize to 0, calculated below
                    "total_cost": 0.0,  # Initialize to 0.0, calculated below
                    "eval_cost": 0.0,  # Initialize to 0.0, calculated below (USD or non-Snowflake)
                    "eval_cost_snowflake": 0.0,  # Initialize to 0.0, calculated below (Snowflake credits)
                    "cost_currency": "USD",  # Initialize to "USD", calculated below
                    "feedback_results": {},  # Initialize to empty map, calculated below
                }

            record_events[record_id]["events"].append(event)

            # Check if the span is of type RECORD_ROOT
            if (
                record_attributes.get(SpanAttributes.SPAN_TYPE)
                == SpanAttributes.SpanType.RECORD_ROOT.value
            ):
                record_events[record_id]["input_id"] = record_attributes.get(
                    SpanAttributes.INPUT_ID, ""
                )
                record_events[record_id]["input"] = record_attributes.get(
                    SpanAttributes.RECORD_ROOT.INPUT, ""
                )
                record_events[record_id]["output"] = record_attributes.get(
                    SpanAttributes.RECORD_ROOT.OUTPUT, ""
                )
                # NOTE: We grab timestamps from the RECORD_ROOT span because it provides a
                # more accurate duration/latency.
                record_events[record_id]["ts"] = event.start_timestamp
                record_events[record_id]["latency"] = (
                    event.timestamp - event.start_timestamp
                ).total_seconds()

            # Check if the span has cost info (tokens, cost, currency)
            # For EVAL/EVAL_ROOT spans, track the cost separately as eval_cost.
            span_type = record_attributes.get(SpanAttributes.SPAN_TYPE)
            if span_type in [
                SpanAttributes.SpanType.EVAL.value,
                SpanAttributes.SpanType.EVAL_ROOT.value,
            ]:
                currency = record_attributes.get(
                    SpanAttributes.COST.CURRENCY, "USD"
                )
                amount = record_attributes.get(SpanAttributes.COST.COST, 0.0)
                if currency == "Snowflake credits":
                    record_events[record_id]["eval_cost_snowflake"] += amount
                else:
                    record_events[record_id]["eval_cost"] += amount
                # Do not add EVAL costs to total_cost to avoid lumping
            else:
                # For non-eval spans, update total_cost and tokens
                self._update_cost_info_otel(
                    record_events[record_id],
                    record_attributes,
                    include_tokens=True,
                )

        # Process feedback results
        feedback_col_names = []
        for record_id, record_data in record_events.items():
            for event in record_data["events"]:
                record_attributes = self._get_event_record_attributes_otel(
                    event
                )

                # Check if the span is of type EVAL or EVAL_ROOT
                if record_attributes.get(SpanAttributes.SPAN_TYPE) in [
                    SpanAttributes.SpanType.EVAL.value,
                    SpanAttributes.SpanType.EVAL_ROOT.value,
                ]:
                    metric_name = record_attributes.get(
                        SpanAttributes.EVAL.METRIC_NAME
                    )
                    if not metric_name:
                        logger.warning(
                            f"Skipping eval span for record_id: {record_id}, no metric name found"
                        )
                        continue

                    # Add feedback name to column names if not present
                    if metric_name not in feedback_col_names:
                        feedback_col_names.append(metric_name)

                    # Initialize feedback result if not present
                    if metric_name not in record_data["feedback_results"]:
                        record_data["feedback_results"][metric_name] = {
                            "mean_score": None,
                            "calls": [],
                            "total_cost": 0.0,
                            "cost_currency": "USD",  # Initialize to USD, calculated below
                            "direction": None,
                        }

                    # Update feedback result
                    # TODO(otel): This isn't going to work if there are multiple with the same name.
                    feedback_result = record_data["feedback_results"][
                        metric_name
                    ]

                    eval_root_score = record_attributes.get(
                        SpanAttributes.EVAL_ROOT.SCORE, None
                    )

                    if (
                        record_attributes.get(SpanAttributes.SPAN_TYPE)
                        == SpanAttributes.SpanType.EVAL_ROOT.value
                    ):
                        # NOTE: EVAL_ROOT.SCORE should provide the mean score of all related EVAL spans
                        feedback_result["mean_score"] = eval_root_score
                        # TODO(SNOW-2112879): HIGHER_IS_BETTER has not been populated in the OTEL spans yet
                        feedback_result["direction"] = record_attributes.get(
                            SpanAttributes.EVAL_ROOT.HIGHER_IS_BETTER,
                            None,
                        )
                        # Add call data for EVAL_ROOT spans
                        args_span_id = self._extract_namespaced_attributes(
                            record_attributes,
                            SpanAttributes.EVAL_ROOT.ARGS_SPAN_ID,
                        )
                        args_span_attribute = (
                            self._extract_namespaced_attributes(
                                record_attributes,
                                SpanAttributes.EVAL_ROOT.ARGS_SPAN_ATTRIBUTE,
                            )
                        )

                        call_data = {
                            "span_type": record_attributes.get(
                                SpanAttributes.SPAN_TYPE
                            ),
                            "eval_root_id": record_attributes.get(
                                SpanAttributes.EVAL.EVAL_ROOT_ID
                            ),
                            "timestamp": record_data["ts"],
                            "args_span_id": args_span_id,
                            "args_span_attribute": args_span_attribute,
                        }
                        feedback_result["calls"].append(call_data)

                        # Update feedback result with cost info if available
                        self._update_cost_info_otel(
                            feedback_result, record_attributes
                        )

                    if (
                        record_attributes.get(SpanAttributes.SPAN_TYPE)
                        == SpanAttributes.SpanType.EVAL.value
                    ):
                        # Extract namespaced attributes using the helper method
                        kwargs = self._extract_namespaced_attributes(
                            record_attributes, SpanAttributes.CALL.KWARGS
                        )

                        call_data = {
                            "span_type": record_attributes.get(
                                SpanAttributes.SPAN_TYPE
                            ),
                            "args": kwargs,
                            "ret": record_attributes.get(
                                SpanAttributes.EVAL.SCORE
                            ),
                            "eval_root_id": record_attributes.get(
                                SpanAttributes.EVAL.EVAL_ROOT_ID
                            ),
                            "timestamp": record_data["ts"],
                            "meta": {
                                "explanation": record_attributes.get(
                                    SpanAttributes.EVAL.EXPLANATION
                                ),
                                "metadata": record_attributes.get(
                                    SpanAttributes.EVAL.METADATA, {}
                                ),
                            },
                        }
                        feedback_result["calls"].append(call_data)

                        # Update feedback result with cost info if available
                        self._update_cost_info_otel(
                            feedback_result, record_attributes
                        )

        # Create dataframe
        records_data = []
        for record_id, record_data in record_events.items():
            # TODO: audit created jsons for correctness (app_json, record_json, cost_json, perf_json)

            app_json = {
                "app_name": record_data["app_name"],
                "app_version": record_data["app_version"],
                "app_id": record_data["app_id"],
            }

            record_json = {
                "record_id": record_id,
                "app_id": record_data["app_id"],
                "input": record_data["input"],
                "output": record_data["output"],
                "tags": record_data["tags"],
                "ts": record_data["ts"],
                "meta": {},
            }

            cost_json = {
                "n_tokens": record_data["total_tokens"],
                # TODO: convert to map (see comment: https://github.com/truera/trulens/pull/1939#discussion_r2054802093)
                "cost": record_data["total_cost"],
            }

            perf_json = {
                "start_time": record_data["ts"],
                "end_time": record_data["ts"]
                + pd.Timedelta(seconds=record_data["latency"]),
            }

            # Create record row
            record_row = {
                "app_id": record_data["app_id"],
                "app_name": record_data["app_name"],
                "app_version": record_data["app_version"],
                "app_json": app_json,
                # TODO(nit): consider using a constant here
                "type": "SPAN",  # Default type as per orm.py
                "record_id": record_id,
                "input_id": record_data.get("input_id"),
                "input": record_data["input"],
                "output": record_data["output"],
                "tags": record_data["tags"],
                "record_json": record_json,
                "cost_json": cost_json,
                "perf_json": perf_json,
                "ts": record_data["ts"],
                "latency": record_data["latency"],
                "total_tokens": record_data["total_tokens"],
                # TODO: convert to map (see comment: https://github.com/truera/trulens/pull/1939#discussion_r2054802093)
                "total_cost": record_data["total_cost"],
                "eval_cost": record_data["eval_cost"],
                "eval_cost_snowflake": record_data["eval_cost_snowflake"],
                "cost_currency": record_data["cost_currency"],
                "num_events": len(record_data["events"]),
            }

            # Add feedback results
            for feedback_name, feedback_result in record_data[
                "feedback_results"
            ].items():
                # NOTE: we use the mean score as the feedback result
                record_row[feedback_name] = feedback_result["mean_score"]

                record_row[f"{feedback_name}_calls"] = feedback_result["calls"]
                record_row[
                    f"{feedback_name} feedback cost in {feedback_result['cost_currency']}"
                ] = feedback_result["total_cost"]
                record_row[f"{feedback_name} direction"] = feedback_result[
                    "direction"
                ]

            records_data.append(record_row)

        # Create dataframe
        df = pd.DataFrame(records_data)

        # Ensure that all expected columns are present
        for col in BaseAppsExtractor.all_cols:
            if col not in df.columns:
                logger.warning(
                    f"Column {col} not found in dataframe, setting to None."
                )
                df[col] = None

        return df, feedback_col_names

    def _extract_namespaced_attributes(
        self, record_attributes: Dict[str, Any], namespace_prefix: str
    ) -> Dict[str, Any]:
        """Extract attributes that are namespaced under a given prefix.

        Args:
            record_attributes: Dictionary of all span attributes
            namespace_prefix: The namespace prefix to look for (e.g., "ai.observability.call.kwargs")

        Returns:
            Dictionary with namespace-stripped keys and their values
        """
        prefix_with_dot = namespace_prefix + "."
        return {
            key[len(prefix_with_dot) :]: value
            for key, value in record_attributes.items()
            if key.startswith(prefix_with_dot)
        }

    def _get_event_record_attributes_otel(self, event: Event) -> Dict[str, Any]:
        """Get the record attributes from the event.

        This implementation differs from the pre-OTEL implementation by using the
        `record_attributes` field of the event.

        Args:
            event: The event to extract the record attributes from.

        Returns:
            Dict[str, Any]: The record attributes from the event.
        """
        record_attributes = event.record_attributes
        if not isinstance(record_attributes, dict):
            try:
                record_attributes = json.loads(record_attributes)
            except (json.JSONDecodeError, TypeError):
                logger.error(
                    f"Failed to decode record attributes as JSON: {record_attributes}",
                )

        return record_attributes

    def _get_event_resource_attributes_otel(
        self, event: Event
    ) -> Dict[str, Any]:
        """Get the resource attributes from the event.

        This implementation differs from the pre-OTEL implementation by using the
        `resource_attributes` field of the event.

        Args:
            event: The event to extract the resource attributes from.

        Returns:
            Dict[str, Any]: The resource attributes from the event.
        """
        resource_attributes = event.resource_attributes
        if not isinstance(resource_attributes, dict):
            try:
                resource_attributes = json.loads(resource_attributes)
            except (json.JSONDecodeError, TypeError):
                logger.error(
                    f"Failed to decode resource attributes as JSON: {resource_attributes}",
                )

        return resource_attributes

    def _update_cost_info_otel(
        self,
        target_dict: dict,
        record_attributes: dict,
        include_tokens: bool = False,
    ):
        """Update cost information in the target dictionary.

        Args:
            target_dict: Dictionary to update with cost information
            record_attributes: Source attributes containing cost information
            include_tokens: Whether to update token count (only for record_events)
        """
        if any(
            key.startswith(SpanAttributes.COST.base)
            for key in record_attributes
        ):
            if include_tokens:
                target_dict["total_tokens"] += record_attributes.get(
                    SpanAttributes.COST.NUM_TOKENS, 0
                )

            target_dict["total_cost"] += record_attributes.get(
                SpanAttributes.COST.COST, 0.0
            )
            target_dict["cost_currency"] = record_attributes.get(
                SpanAttributes.COST.CURRENCY, "USD"
            )

        # TODO(SNOW-2061174): convert to map (see comment: https://github.com/truera/trulens/pull/1939#discussion_r2054802093)
        # Add to total_cost map
        # cost = record_attributes.get(SpanAttributes.COST.COST, 0.0)
        # currency = record_attributes.get(SpanAttributes.COST.CURRENCY, "USD")
        # if currency not in record_events[record_id]["total_cost"]:
        #     record_events[record_id]["total_cost"][currency] = 0.0
        # record_events[record_id]["total_cost"][currency] += cost

    @staticmethod
    def extract_app_and_run_info(
        attributes: Dict[str, Any], resource_attributes: Dict[str, Any]
    ) -> Tuple[str, str, str, str]:
        """Get app info from attributes.

        Args:
            attributes: Span attributes of record root.
            resource_attributes: Resource attributes of record root.

        Returns:
            Tuple of: app name, app version, app id, and run name.
        """

        def get_value(keys: List[str]) -> Optional[str]:
            for key in keys:
                for attr in [resource_attributes, attributes]:
                    if key in attr:
                        return attr[key]
            return None

        app_name = get_value([
            f"snow.{BASE_SCOPE}.object.name",
            ResourceAttributes.APP_NAME,
        ])
        app_version = (
            get_value([
                f"snow.{BASE_SCOPE}.object.version.name",
                ResourceAttributes.APP_VERSION,
            ])
            or "base"
        )
        app_id = get_value([ResourceAttributes.APP_ID])
        if app_id is None:
            app_id = app_schema.AppDefinition._compute_app_id(
                app_name, app_version
            )
        run_name = get_value([
            f"snow.{BASE_SCOPE}.run.name",
            SpanAttributes.RUN_NAME,
        ])
        return app_name, str(app_version), app_id, run_name



"""Type aliases."""

from typing import Dict
import uuid

from trulens.core._utils.pycompat import TypeAlias  # import style exception

RecordID: TypeAlias = str
"""Unique identifier for a record.

By default these hashes of record content as json.
[Record.record_id][trulens.core.schema.record.Record.record_id].
"""

CallID: TypeAlias = str
"""Unique identifier for a record app call.

See [RecordAppCall.call_id][trulens.core.schema.record.RecordAppCall.call_id].
"""


def new_call_id() -> CallID:
    """Generate a new call id."""
    return str(uuid.uuid4())


AppID: TypeAlias = str
"""Unique identifier for an app.

By default these are hashes of app content as json.
See [AppDefinition.app_id][trulens.core.schema.app.AppDefinition.app_id].
"""

AppName: TypeAlias = str
"""Unique App name.

See [AppDefinition.app_name][trulens.core.schema.app.AppDefinition.app_name].
"""

AppVersion: TypeAlias = str
"""Version identifier for an app.

See [AppDefinition.app_version][trulens.core.schema.app.AppDefinition.app_version].
"""

RunName: TypeAlias = str
"""Run name."""

Tags: TypeAlias = str
"""Tags for an app or record.

See [AppDefinition.tags][trulens.core.schema.app.AppDefinition.tags] and
[Record.tags][trulens.core.schema.record.Record.tags].
"""

Metadata: TypeAlias = Dict
"""Metadata for an app, record, groundtruth, or dataset.

See [AppDefinition.metadata][trulens.core.schema.app.AppDefinition.metadata],
[Record.meta][trulens.core.schema.record.Record.meta], [GroundTruth.meta][trulens.core.schema.groundtruth.GroundTruth.meta], and
[Dataset.meta][trulens.core.schema.dataset.Dataset.meta].
"""

FeedbackDefinitionID: TypeAlias = str
"""Unique identifier for a feedback definition.

By default these are hashes of feedback definition content as json. See
[FeedbackDefinition.feedback_definition_id][trulens.core.schema.feedback.FeedbackDefinition.feedback_definition_id].
"""

FeedbackResultID: TypeAlias = str
"""Unique identifier for a feedback result.

By default these are hashes of feedback result content as json. See
[FeedbackResult.feedback_result_id][trulens.core.schema.feedback.FeedbackResult].
"""

GroundTruthID: TypeAlias = str
"""Unique identifier for a groundtruth.

By default these are hashes of ground truth content as json.
"""

DatasetID: TypeAlias = str
"""Unique identifier for a dataset.

By default these are hashes of dataset content as json.
"""

EventID: TypeAlias = str
"""Unique identifier for a event.
"""



"""Basic input output instrumentation and monitoring."""

from inspect import BoundArguments
from inspect import Signature
from inspect import signature
import logging
from pprint import PrettyPrinter
from typing import Any, Callable, ClassVar, List, Optional

from pydantic import Field
from trulens.core import app as core_app
from trulens.core import instruments as core_instruments
from trulens.core.instruments import InstrumentedMethod
from trulens.core.utils import pyschema as pyschema_utils

logger = logging.getLogger(__name__)

pp = PrettyPrinter()


class TruWrapperApp:
    """Wrapper of basic apps.

    This will be wrapped by instrumentation.

    Warning:
        Because `TruWrapperApp` may wrap different types of callables, we cannot
        patch the signature to anything consistent. Because of this, the
        dashboard/record for this call will have `*args`, `**kwargs` instead of
        what the app actually uses. We also need to adjust the main_input lookup
        to get the correct signature. See note there.
    """

    def _call(self, *args, **kwargs):
        return self._call_fn(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self._call(*args, **kwargs)

    def __init__(self, call_fn: Callable):
        self._call_fn = call_fn


class TruBasicCallableInstrument(core_instruments.Instrument):
    """Basic app instrumentation."""

    class Default:
        """Default instrumentation specification for basic apps."""

        CLASSES = lambda: {TruWrapperApp}

        # Instrument only methods with these names and of these classes.
        METHODS: List[InstrumentedMethod] = [
            InstrumentedMethod("_call", TruWrapperApp)
        ]

    def __init__(self, *args, **kwargs):
        super().__init__(
            include_classes=TruBasicCallableInstrument.Default.CLASSES(),
            include_methods=TruBasicCallableInstrument.Default.METHODS,
            *args,
            **kwargs,
        )


class TruBasicApp(core_app.App):
    """Instantiates a Basic app that makes little assumptions.

    Assumes input text and output text.

    Example:
        ```python
        def custom_application(prompt: str) -> str:
            return "a response"

        from trulens.apps.basic import TruBasicApp
        # f_lang_match, f_qa_relevance, f_context_relevance are feedback functions
        tru_recorder = TruBasicApp(custom_application,
            app_name="Custom Application",
            app_version="1",
            feedbacks=[f_lang_match, f_qa_relevance, f_context_relevance])

        # Basic app works by turning your callable into an app
        # This app is accessible with the `app` attribute in the recorder
        with tru_recorder as recording:
            tru_recorder.app(question)

        tru_record = recording.records[0]
        ```

        See [Feedback
        Functions](https://www.trulens.org/trulens/api/feedback/) for
        instantiating feedback functions.

    Args:
        text_to_text: A str to str callable.

        app: A TruWrapperApp instance. If not provided, `text_to_text` must
            be provided.

        **kwargs: Additional arguments to pass to [App][trulens.core.app.App]
            and [AppDefinition][trulens.core.schema.app.AppDefinition]
    """

    app: TruWrapperApp
    """The app to be instrumented."""

    # TODEP
    root_callable: ClassVar[pyschema_utils.FunctionOrMethod] = Field(None)
    """The root callable to be instrumented.

    This is the method that will be called by the main_input method."""

    def __init__(
        self,
        text_to_text: Optional[Callable[[str], str]] = None,
        app: Optional[TruWrapperApp] = None,
        **kwargs: Any,
    ):
        if text_to_text is not None:
            app = TruWrapperApp(text_to_text)
        else:
            assert (
                app is not None
            ), "Need to provide either `app: TruWrapperApp` or a `text_to_text: Callable`."
        if "main_method" in kwargs and kwargs["main_method"] is not None:
            raise ValueError(
                "`main_method` should not be provided for `TruBasicApp`!"
            )

        kwargs["app"] = app
        kwargs["root_class"] = pyschema_utils.Class.of_object(app)
        kwargs["instrument"] = TruBasicCallableInstrument(app=self)
        kwargs["main_method"] = app._call

        super().__init__(**kwargs)

    def main_call(self, human: str) -> str:
        # If available, a single text to a single text invocation of this app.

        return self.app._call(human)

    def main_input(
        self, func: Callable, sig: Signature, bindings: BoundArguments
    ) -> str:
        if hasattr(
            TruWrapperApp._call, core_instruments.Instrument.INSTRUMENT
        ) and func == getattr(
            TruWrapperApp._call, core_instruments.Instrument.INSTRUMENT
        ):
            # If func is the wrapper app _call, replace the signature and
            # bindings based on the actual containing callable instead of
            # self.app._call . This needs to be done since the a TruWrapperApp
            # may be wrapping apps with different signatures on their callables
            # so TruWrapperApp._call cannot have a consistent signature
            # statically. Note also we are looking up the Instrument.INSTRUMENT
            # attribute here since the method is instrumented and overridden by
            # another wrapper in the process with the original accessible at
            # this attribute.

            sig = signature(self.app._call_fn)
            # Skipping self as TruWrapperApp._call takes in self, but
            # self.app._call_fn does not.
            bindings = sig.bind(*bindings.args[1:], **bindings.kwargs)

        return super().main_input(func, sig, bindings)

    def call_with_record(self, *args, **kwargs) -> None:
        self._throw_dep_message(method="call", is_async=False, with_record=True)


TruBasicApp.model_rebuild()



"""# Serializable Classes

Note: Only put classes which can be serialized in this module.

# Classes with non-serializable variants

Many of the classes defined here extending serial.SerialModel are meant to be
serialized into json. Most are extended with non-serialized fields in other files.

| Serializable       | Non-serializable        |
| ------------------ | ----------------------- |
| [AppDefinition][trulens.core.schema.app.AppDefinition] | [App][trulens.core.app.App], Tru{Chain, Llama, ...} |
| [FeedbackDefinition][trulens.core.schema.feedback.FeedbackDefinition] | [Feedback][trulens.core.Feedback] |

`AppDefinition.app` is the JSON-ized version of a wrapped app while `App.app` is the
actual wrapped app. We can thus inspect the contents of a wrapped app without
having to construct it. Additionally, JSONized objects like `AppDefinition.app`
feature information about the encoded object types in the dictionary under the
`core/utils/constantx.py:CLASS_INFO` key.
"""

# WARNING: This file does not follow the no-init aliases import standard.

from trulens.core.schema.app import AppDefinition
from trulens.core.schema.dataset import Dataset
from trulens.core.schema.feedback import FeedbackDefinition
from trulens.core.schema.feedback import FeedbackMode
from trulens.core.schema.feedback import FeedbackResult
from trulens.core.schema.groundtruth import GroundTruth
from trulens.core.schema.record import Record
from trulens.core.schema.select import Select

__all__ = [
    "AppDefinition",
    "Select",
    "FeedbackDefinition",
    "FeedbackResult",
    "FeedbackMode",
    "Record",
    "GroundTruth",
    "Dataset",
]



from __future__ import annotations  # defers evaluation of annotations

from enum import Enum
import inspect
import json
import logging
import time
from typing import Any, ClassVar, Dict, List, Optional, Set, Type, Union

import pandas as pd
import pydantic
from pydantic import BaseModel
from pydantic import Field
from pydantic import field_serializer
from trulens.core.enums import Mode
from trulens.core.feedback.custom_metric import MetricConfig
from trulens.core.metric import metric as metric_module
from trulens.core.utils.json import obj_id_of_obj
from trulens.otel.semconv.trace import SpanAttributes

logger = logging.getLogger(__name__)


def _parse_json_field(
    value: Any,
    field_name: str,
    index: Optional[int] = None,
    critical: bool = True,
) -> Any:
    """
    Helper function to parse JSON fields consistently.

    Args:
        value: The value to parse (could be string or already parsed dict)
        field_name: Name of the field being parsed (for logging)
        index: Optional index for logging context
        critical: Whether parsing failure should cause the caller to continue processing

    Returns:
        Parsed dictionary or the original value if already parsed

    Raises:
        Continues processing on JSONDecodeError if critical=False, otherwise logs warning
    """
    if isinstance(value, str):
        try:
            return json.loads(value)
        except json.JSONDecodeError:
            index_msg = f" at index {index}" if index is not None else ""
            if critical:
                logger.warning(f"Failed to parse {field_name} JSON{index_msg}")
                raise
            else:
                logger.debug(f"Failed to parse {field_name} JSON{index_msg}")
                return value
    return value


def _get_all_span_attribute_key_constants(cls: Type, prefix: str) -> List[str]:
    ret = []
    for curr_name in dir(cls):
        if (not curr_name.startswith("__")) and (not curr_name.endswith("__")):
            curr = getattr(cls, curr_name)
            if inspect.isclass(curr):
                ret += _get_all_span_attribute_key_constants(
                    curr, f"{curr_name}"
                )
            elif curr_name == curr_name.upper():
                ret += [f"{prefix + '.' if prefix else ''}{curr_name}"]
    ret = list(set(ret))
    ret = [field for field in ret if not (field.startswith("SpanType"))]

    ret += [
        field.replace(
            "RECORD_ROOT.", ""
        )  # record_root can be optionally specified
        for field in ret
        if field.startswith("RECORD_ROOT.")
    ]
    return ret


def get_all_span_attribute_key_constants() -> set[str]:
    return set(_get_all_span_attribute_key_constants(SpanAttributes, ""))


# Reserved fields (case-insensitive) for dataset specification directly maps to OTEL Span attributes
DATASET_RESERVED_FIELDS: Set[str] = {
    field.lower() for field in get_all_span_attribute_key_constants()
}


EXPECTED_TELEMETRY_LATENCY_IN_MS = (
    2 * 60 * 1000
)  # expected latency from the telemetry pipeline before ingested rows show up in event table


class RunStatus(str, Enum):
    # note this the inferred / computeted status determined by SDK, and is different from the DPO entity level run_status

    # invocation statuses
    INVOCATION_IN_PROGRESS = "INVOCATION_IN_PROGRESS"
    INVOCATION_COMPLETED = "INVOCATION_COMPLETED"
    INVOCATION_PARTIALLY_COMPLETED = "INVOCATION_PARTIALLY_COMPLETED"

    # computation statuses
    COMPUTATION_IN_PROGRESS = "COMPUTATION_IN_PROGRESS"
    COMPLETED = "COMPLETED"
    PARTIALLY_COMPLETED = "PARTIALLY_COMPLETED"

    CREATED = "CREATED"
    FAILED = "FAILED"
    CANCELLED = "CANCELLED"
    UNKNOWN = "UNKNOWN"


class SupportedEntryType(str, Enum):
    INVOCATIONS = "invocations"
    COMPUTATIONS = "computations"
    METRICS = "metrics"


SUPPORTED_ENTRY_TYPES = [e.value for e in SupportedEntryType]


def validate_dataset_spec(
    dataset_spec: Dict[str, str],
) -> Dict[str, str]:
    """
    Validates and normalizes the dataset column specification to ensure it contains only
    currently supported span attributes and that the keys are in the correct format.

    Args:
        dataset_spec: The user-provided dictionary with column names.

    Returns:
        A validated and normalized dictionary.

    Raises:
        ValueError: If any invalid field is present.
    """

    normalized_spec = {}

    for key, value in dataset_spec.items():
        normalized_key = key.lower()
        # Add the normalized field to the dictionary
        normalized_spec[normalized_key] = value

    return normalized_spec


class RunConfig(BaseModel):
    run_name: str = Field(
        ...,
        description="Unique name of the run. This name should be unique within the object.",
    )
    dataset_name: str = Field(
        default=...,
        description="Mandatory field. The name of a user's Snowflake Table / View  (e.g. 'user_table_name_1'), or any user specified name of input dataframe.",
    )

    source_type: str = Field(
        default="DATAFRAME",
        description="Type of the source (e.g. 'DATAFRAME' for user provided dataframe or 'TABLE' for user table in Snowflake).",
    )

    dataset_spec: Dict[str, str] = Field(
        default=...,
        description="Mandatory column name mapping from reserved dataset fields to column names in user's table.",
    )

    description: Optional[str] = Field(
        default=None, description="A description for the run."
    )
    label: Optional[str] = Field(
        default=None,
        description="Text label to group the runs. Take a single label for now",
    )
    llm_judge_name: Optional[str] = Field(
        default=None,
        description="Name of the LLM judge to be used for the run.",
    )
    mode: Mode = Field(
        default=Mode.APP_INVOCATION,
        description="Mode of operation: LOG_INGESTION for creating spans from existing data, APP_INVOCATION for instrumenting spans from a new app execution.",
    )


class Run(BaseModel):
    model_config: ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(
        arbitrary_types_allowed=True,
        extra="ignore",  # allow custom obj like RunDao to be passed as a parameter and more importantly, account for
        # additional fields in Run metadata JSON response.
    )

    """
    Run class for managing run state / attributes in the SDK client.

    This model is meant to be used and accessed through
    methods like describe() (which uses the underlying RunDao) to obtain the run metadata.
    """

    run_dao: Any = Field(
        ..., description="DAO instance for run operations.", exclude=True
    )

    app: Any = Field(
        ...,
        description="TruLens app/recorder instance to be invoked during run.",
        exclude=True,
    )

    main_method_name: Optional[str] = Field(
        default=None, description="Main method of the app.", exclude=True
    )

    tru_session: Any = Field(
        ..., description="TruSession instance.", exclude=True
    )
    object_name: str = Field(
        ...,
        description="Name of the managing object (e.g. name of 'EXTERNAL AGENT').",
    )

    object_type: str = Field(
        ..., description="Type of the managing object (e.g. 'EXTERNAL AGENT')."
    )

    object_version: Optional[str] = Field(
        default=None, description="Version of the managing object."
    )

    run_name: str = Field(
        ...,
        description="Unique name of the run. This name should be unique within the object.",
    )

    run_status: Optional[str] = Field(
        default=None,
        description="The status of the run on the entity level. Currently it can only be ACTIVE or CANCELLED.",
    )

    description: Optional[str] = Field(
        default=None, description="A description for the run."
    )

    class RunMetadata(BaseModel):
        labels: Optional[List[str]] = Field(
            default=None,
            description="Text label to group the runs. Take a single label for now",
        )
        llm_judge_name: Optional[str] = Field(
            default=None,
            description="Name of the LLM judge to be used for the run.",
        )
        mode: Optional[str] = Field(
            default=Mode.APP_INVOCATION.value,
            description="Mode of operation: LOG_INGESTION or APP_INVOCATION.",
        )
        invocations: Optional[Dict[str, Run.InvocationMetadata]] = Field(
            default=None,
            description="Map of invocation metadata with invocation ID as key.",
        )
        computations: Optional[Dict[str, Run.ComputationMetadata]] = Field(
            default=None,
            description="Map of computation metadata with computation ID as key.",
        )
        metrics: Optional[Dict[str, Run.MetricsMetadata]] = Field(
            default=None,
            description="Map of metrics metadata with metric ID as key.",
        )

    run_metadata: RunMetadata = Field(
        default=...,
        description="Run metadata that maintains states needed for app invocation and metrics computation.",
    )

    class SourceInfo(BaseModel):
        name: str = Field(
            ...,
            description="Name of the source (e.g. name of the table).",
        )
        column_spec: Dict[str, str] = Field(
            default=...,
            description="Column name mapping from reserved dataset fields to column names in user's table.",
        )
        source_type: str = Field(
            default="DATAFRAME",
            description="Type of the source (e.g. 'DATAFRAME').",
        )

    source_info: SourceInfo = Field(
        default=...,
        description="Source information for the run.",
    )

    class CompletionStatusStatus(str, Enum):
        UNKNOWN = "UNKNOWN"
        STARTED = "STARTED"
        PARTIALLY_COMPLETED = "PARTIALLY_COMPLETED"
        COMPLETED = "COMPLETED"
        FAILED = "FAILED"

    class CompletionStatus(BaseModel):
        status: Run.CompletionStatusStatus = Field(
            ..., description="The status of the completion."
        )
        record_count: Optional[int] = Field(
            default=None, description="The count of records processed."
        )

        @field_serializer("status")
        def serialize_status(self, status: Run.CompletionStatusStatus, _info):
            return status.value

    class InvocationMetadata(BaseModel):
        input_records_count: Optional[int] = Field(
            default=None,
            description="The number of input records in the dataset.",
        )
        id: Optional[str] = Field(
            default=None,
            description="The unique identifier for the invocation metadata.",
        )
        start_time_ms: Optional[int] = Field(
            default=None,
            description="The start time of the invocation.",
        )
        end_time_ms: Optional[int] = Field(
            default=None,
            description="The end time of the invocation.",
        )
        completion_status: Optional[Run.CompletionStatus] = Field(
            default=None,
            description="The status of the invocation.",
        )

    class ComputationMetadata(BaseModel):
        id: Optional[str] = Field(
            default=None,
            description="Unique id, even if name is repeated.",
        )
        query_id: Optional[str] = Field(
            default=None,
            description="Query id associated with metric computation.",
        )
        start_time_ms: Optional[int] = Field(
            default=None,
            description="Start time of the computation in milliseconds.",
        )
        end_time_ms: Optional[int] = Field(
            default=None,
            description="End time of the computation in milliseconds.",
        )

    class MetricsMetadata(BaseModel):
        id: Optional[str] = Field(
            default=None,
            description="Unique id for the metrics metadata.",
        )
        name: Optional[str] = Field(
            default=None,
            description="Name of the metric.",
        )
        completion_status: Optional[Run.CompletionStatus] = Field(
            default=None,
            description="Completion status of the metric computation.",
        )
        computation_id: Optional[str] = Field(
            default=None,
            description="ID of the computation associated with the metric.",
        )

    def describe(self) -> dict:
        """
        Retrieve the metadata of the Run object.
        """

        run_metadata_df = self.run_dao.get_run(
            run_name=self.run_name,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
        )
        if run_metadata_df.empty:
            raise ValueError(f"Run {self.run_name} not found.")

        raw_json = json.loads(
            list(run_metadata_df.to_dict(orient="records")[0].values())[0]
        )

        # remove / hide entity-level run_status field to avoid customer's confusion
        raw_json.pop("run_status", None)
        return raw_json

    def delete(self) -> None:
        """
        Delete the run by its name and object name.
        """
        self.run_dao.delete_run(
            run_name=self.run_name,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
        )

    def _can_start_new_invocation(self, current_run_status: RunStatus) -> bool:
        """
        Check if the run is in a state that allows starting a new invocation.
        """
        if self._is_cancelled():
            logger.warning("Cannot start a new invocation for a cancelled run.")
            return False

        return current_run_status in [
            RunStatus.CREATED,
            RunStatus.INVOCATION_PARTIALLY_COMPLETED,
            RunStatus.FAILED,
            RunStatus.UNKNOWN,
        ]

    def _can_start_new_metric_computation(
        self, current_run_status: RunStatus
    ) -> bool:
        """
        Check if the run is in a state that allows starting a new metric computation.
        """
        if self._is_cancelled():
            logger.warning(
                "Cannot start a new metric computation for a cancelled run."
            )
            return False

        if current_run_status == RunStatus.COMPUTATION_IN_PROGRESS:
            logger.warning(
                "Previous computation(s) still in progress. Starting another new metric computation when computation is in progress."
            )

        return current_run_status in [
            RunStatus.INVOCATION_COMPLETED,
            RunStatus.INVOCATION_PARTIALLY_COMPLETED,
            RunStatus.COMPUTATION_IN_PROGRESS,
            RunStatus.COMPLETED,
            RunStatus.PARTIALLY_COMPLETED,
            RunStatus.FAILED,
        ]

    def _is_invocation_started(self, run: Run) -> bool:
        return (
            run.run_metadata.invocations is not None
            and len(run.run_metadata.invocations) > 0
        )

    def _compute_latest_invocation_status(self, run: Run) -> RunStatus:
        latest_invocation = max(
            run.run_metadata.invocations.values(),
            key=lambda inv: (inv.start_time_ms or 0, inv.id or ""),
        )

        if (
            latest_invocation.completion_status
            and latest_invocation.completion_status.status
        ):
            completion_status = latest_invocation.completion_status.status
            if completion_status == Run.CompletionStatusStatus.COMPLETED:
                return RunStatus.INVOCATION_COMPLETED
            elif (
                completion_status
                == Run.CompletionStatusStatus.PARTIALLY_COMPLETED
            ):
                return RunStatus.INVOCATION_PARTIALLY_COMPLETED
            elif completion_status == Run.CompletionStatusStatus.STARTED:
                return RunStatus.INVOCATION_IN_PROGRESS
            elif completion_status == Run.CompletionStatusStatus.FAILED:
                return RunStatus.FAILED
            else:
                logger.warning(
                    f"Unknown completion status {completion_status} for invocation {latest_invocation.id}"
                )
                return RunStatus.UNKNOWN

    def _metrics_computation_started(self, run: Run) -> bool:
        return (
            run.run_metadata.metrics is not None
            and len(run.run_metadata.metrics) > 0
        )

    def _resolve_overall_metrics_status(
        self, all_metrics, invocation_completion_status
    ):
        """
        Given the list of all metrics, resolve and return the overall metrics status.
        """
        if all(
            metric.completion_status
            and metric.completion_status.status
            == Run.CompletionStatusStatus.COMPLETED
            for metric in all_metrics
        ):
            return (
                RunStatus.COMPLETED
                if invocation_completion_status
                == Run.CompletionStatusStatus.COMPLETED
                else RunStatus.PARTIALLY_COMPLETED
            )
        elif all(
            metric.completion_status
            and metric.completion_status.status
            == Run.CompletionStatusStatus.FAILED
            for metric in all_metrics
        ):
            return RunStatus.FAILED
        elif all(
            metric.completion_status
            and metric.completion_status.status
            in [
                Run.CompletionStatusStatus.COMPLETED,
                Run.CompletionStatusStatus.FAILED,
            ]
            for metric in all_metrics
        ):
            return RunStatus.PARTIALLY_COMPLETED
        elif any(
            metric.completion_status.status
            == Run.CompletionStatusStatus.STARTED
            for metric in all_metrics
        ):
            return RunStatus.COMPUTATION_IN_PROGRESS
        else:
            logger.warning(
                "Cannot determine run status. Metrics: %s",
                [metric.name for metric in all_metrics],
            )
            return RunStatus.UNKNOWN

    def _compute_overall_computations_status(self, run: Run) -> RunStatus:
        all_existing_metrics = run.run_metadata.metrics.values()

        latest_invocation = max(
            run.run_metadata.invocations.values(),
            key=lambda inv: (inv.start_time_ms or 0, inv.id or ""),
        )
        invocation_completion_status = (
            latest_invocation.completion_status.status
        )

        # check all metrics and see if their completion status are set or not
        metrics_status_not_set = [
            metric
            for metric in all_existing_metrics
            if not metric.completion_status
            or not metric.completion_status.status
        ]

        if len(metrics_status_not_set) != 0:
            logger.warning(
                f"Metrics status not set for: {[metric.name for metric in metrics_status_not_set]}."
            )
            raise ValueError(
                f"Metrics status not set for: {[metric.name for metric in metrics_status_not_set]}."
            )

        logger.info("All metrics statuses are set as expected.")
        return self._resolve_overall_metrics_status(
            all_existing_metrics, invocation_completion_status
        )

    def start(self, input_df: Optional[pd.DataFrame] = None):
        """
        Start the run by invoking the main method of the user's app with the input data

        Args:
            input_df (Optional[pd.DataFrame], optional): user provided input dataframe.
        """
        current_status = self.get_status()
        logger.info(f"Current run status: {current_status}")
        if not self._can_start_new_invocation(current_status):
            return f"Cannot start a new invocation when in run status: {current_status}. Valid statuses are: {RunStatus.CREATED}, {RunStatus.INVOCATION_PARTIALLY_COMPLETED}, or {RunStatus.FAILED}."

        if input_df is None:
            logger.info(
                "No input dataframe provided. Fetching input data from source."
            )
            rows = self.run_dao.session.sql(
                f"SELECT * FROM {self.source_info.name}"
            ).collect()
            input_df = pd.DataFrame([row.as_dict() for row in rows])

        dataset_spec = self.source_info.column_spec

        input_records_count = len(input_df)

        logger.info(
            f"Creating or updating invocation metadata with {input_records_count} records from input."
        )

        # Determine mode from run metadata
        mode = self.run_metadata.mode or Mode.APP_INVOCATION.value

        try:
            if mode == Mode.LOG_INGESTION.value:
                self._create_virtual_spans(
                    input_df, dataset_spec, input_records_count
                )
            else:
                # user app invocation - will block until the app completes
                for _, row in input_df.iterrows():
                    main_method_args = []

                    # Call the instrumented main method with the arguments
                    # TODO (dhuang) better way to check span attributes, also is this all we need to support?
                    input_id = (
                        row[dataset_spec["input_id"]]
                        if "input_id" in dataset_spec
                        else None
                    )
                    input_col = None
                    if input_id is None:
                        if "input" in dataset_spec:
                            input_col = dataset_spec["input"]
                        elif "record_root.input" in dataset_spec:
                            input_col = dataset_spec["record_root.input"]
                        if input_col:
                            input_value = row[input_col]
                            # Parse JSON string to dict if the input is a JSON blob
                            # This supports LangGraph state dicts from Snowflake VARIANT columns
                            if isinstance(input_value, str):
                                try:
                                    import json

                                    parsed = json.loads(input_value)
                                    if isinstance(parsed, dict):
                                        input_value = parsed
                                except (json.JSONDecodeError, TypeError):
                                    pass  # Keep as string if not valid JSON
                            input_id = obj_id_of_obj(input_value)
                            main_method_args.append(input_value)

                    # Extract additional method arguments from dataset_spec
                    # Skip fields that are already handled or are special metadata fields
                    special_fields = {
                        "input_id",
                        "input",
                        "record_root.input",
                        "ground_truth_output",
                        "record_root.ground_truth_output",
                    }

                    for spec_key, column_name in dataset_spec.items():
                        if (
                            spec_key not in special_fields
                            and column_name in row
                        ):
                            main_method_args.append(row[column_name])

                    ground_truth_output = row.get(
                        dataset_spec.get("ground_truth_output")
                        or dataset_spec.get("record_root.ground_truth_output")
                    )

                    self.app.instrumented_invoke_main_method(
                        run_name=self.run_name,
                        input_id=input_id,
                        input_records_count=input_records_count,
                        ground_truth_output=ground_truth_output,
                        main_method_args=tuple(
                            main_method_args
                        ),  # Ensure correct order
                        main_method_kwargs=None,  # don't take any kwargs for now so we don't break TruChain / TruLlama where input argument name cannot be defined by users.
                    )
        except Exception as e:
            logger.exception(
                f"Error encountered during invoking app main method: {e}."
            )
            raise

        self.tru_session.force_flush()
        logger.info(
            f"Flushed all spans for the run {self.run_name}; and exported them to the telemetry pipeline."
        )

        # we start the ingestion sproc after the app invocation is done, so that
        # app invocation time does not count toward the ingestion timeout set on the task orchestration layer.
        self.run_dao.start_ingestion_query(
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
            run_name=self.run_name,
            input_records_count=input_records_count,
        )
        logger.info("Run started, invocation done and ingestion in process.")

    def _create_virtual_spans(
        self,
        input_df: pd.DataFrame,
        dataset_spec: Dict[str, str],
        input_records_count: int,
    ):
        """
        Create OTEL spans from existing data without actual app invocation.
        This method creates spans dynamically based on the dataset_spec mapping,
        which maps span attribute paths to column names.

        The input DataFrame must conform to a specific format based on the dataset_spec.
        Required columns/data:
        - If 'input_id' is specified in dataset_spec: A column containing unique identifiers
        - If no 'input_id': Either 'input' or 'record_root.input' column must exist to generate IDs
        - If neither exists, row indices will be used as fallback IDs

        Optional columns (specified via dataset_spec):
        - 'ground_truth_output' or 'record_root.ground_truth_output': Ground truth data
        - Any additional columns mapped in dataset_spec will be used to populate span attributes

        See RECORD_ROOT and other span attribute definitions in trulens.otel.semconv.trace
        for the full set of supported attribute paths that can be mapped to columns.
        """

        logger.debug(f"Creating virtual spans for {len(input_df)} records")

        for i, row in input_df.iterrows():
            input_id = (
                row[dataset_spec["input_id"]]
                if "input_id" in dataset_spec
                else None
            )
            input_col = None
            if input_id is None:
                if "input" in dataset_spec:
                    input_col = dataset_spec["input"]
                elif "record_root.input" in dataset_spec:
                    input_col = dataset_spec["record_root.input"]
                if input_col:
                    input_id = obj_id_of_obj(row[input_col])

            ground_truth_output = row.get(
                dataset_spec.get("ground_truth_output")
                or dataset_spec.get("record_root.ground_truth_output")
            )

            # Ensure input_id is never None - use row index as fallback
            if input_id is None:
                input_id = f"row_{i}"
                logger.warning(f"input_id was None, using fallback: {input_id}")

            try:
                self._create_virtual_spans_with_nested_contexts(
                    row,
                    dataset_spec,
                    input_id,
                    input_records_count,
                    ground_truth_output,
                )
            except Exception as e:
                logger.exception(f"Error in virtual span creation: {e}")
                raise

    def _create_virtual_spans_with_nested_contexts(
        self,
        row,
        dataset_spec,
        input_id,
        input_records_count,
        ground_truth_output,
    ):
        """Create virtual spans using OtelRecordingContext - simplified approach"""
        from trulens.core.otel.instrument import OtelRecordingContext

        with OtelRecordingContext(
            tru_app=self.app,
            app_name=self.app.app_name,
            app_version=self.app.app_version,
            run_name=self.run_name,
            input_id=input_id,
            input_records_count=input_records_count,
            ground_truth_output=ground_truth_output,
        ):
            # Now create nested spans within this context
            self._create_nested_spans_from_dataset_spec(dataset_spec, row)

            logger.debug("Created all nested spans")

    def _create_nested_spans_from_dataset_spec(
        self, dataset_spec: Dict[str, str], row
    ):
        """Create properly nested spans within OtelRecordingContext"""
        from trulens.experimental.otel_tracing.core.span import (
            set_general_span_attributes,
        )

        span_data = self._group_dataset_spec_by_span_type(dataset_spec, row)
        logger.info(f"Creating nested spans for: {list(span_data.keys())}")

        if "record_root" in span_data:
            from trulens.core.otel.function_call_context_manager import (
                create_function_call_context_manager,
            )

            with create_function_call_context_manager(
                True, "virtual_record"
            ) as root_span:
                set_general_span_attributes(
                    root_span, SpanAttributes.SpanType.RECORD_ROOT
                )
                self._set_span_attributes_from_data(
                    root_span, span_data["record_root"], "record_root"
                )

                root_span_id = root_span.get_span_context().span_id
                logger.debug(
                    f"Created record_root span with ID: {root_span_id}"
                )

                # Create child spans nested WITHIN the root span context
                for span_type, attributes in span_data.items():
                    if span_type == "record_root":
                        continue

                    span_type_enum = self._get_span_type_enum(span_type)
                    if span_type_enum:
                        logger.debug(
                            f"Creating child span: {span_type} under parent {root_span_id}"
                        )
                        with create_function_call_context_manager(
                            True, f"{span_type}_virtual"
                        ) as child_span:
                            set_general_span_attributes(
                                child_span, span_type_enum
                            )
                            self._set_span_attributes_from_data(
                                child_span, attributes, span_type
                            )

                            child_span_id = (
                                child_span.get_span_context().span_id
                            )
                            child_parent_span_id = (
                                child_span.parent.span_id
                                if child_span.parent
                                else "None"
                            )
                            logger.debug(
                                f"Created {span_type} child span ID: {child_span_id}, parent ID: {child_parent_span_id}"
                            )
                    else:
                        logger.warning(
                            f"Unknown span type: {span_type} - skipping"
                        )
        else:
            logger.warning("No record_root defined in dataset_spec")

    def _set_span_attributes_from_data(
        self, span, attributes: Dict[str, any], span_type: str
    ):
        """Set span attributes using proper SpanAttributes constants"""
        span_attrs_class = self._get_span_attributes_class(span_type)

        for attr_name, value in attributes.items():
            if span_attrs_class:
                attr_constant = getattr(
                    span_attrs_class, attr_name.upper(), None
                )
                if attr_constant:
                    # Check if this attribute should be treated as an array
                    if self._should_process_as_array(attr_constant):
                        processed_value = self._process_array_attribute(value)
                        span.set_attribute(attr_constant, processed_value)
                        logger.info(
                            f"Set {span_type} attribute {attr_constant} = {processed_value}"
                        )
                    else:
                        span.set_attribute(attr_constant, str(value))
                        logger.info(
                            f"Set {span_type} attribute {attr_constant} = {value}"
                        )
                else:
                    # Fallback to raw attribute name
                    fallback_key = f"{span_type}.{attr_name}"
                    # Check if this attribute name suggests it should be an array
                    if self._should_process_as_array_by_name(attr_name):
                        processed_value = self._process_array_attribute(value)
                        span.set_attribute(fallback_key, processed_value)
                        logger.info(
                            f"Set fallback attribute {fallback_key} = {processed_value}"
                        )
                    else:
                        span.set_attribute(fallback_key, str(value))
                        logger.info(
                            f"Set fallback attribute {fallback_key} = {value}"
                        )
            else:
                fallback_key = f"{span_type}.{attr_name}"
                # Check if this attribute name suggests it should be an array
                if self._should_process_as_array_by_name(attr_name):
                    processed_value = self._process_array_attribute(value)
                    span.set_attribute(fallback_key, processed_value)
                    logger.info(
                        f"Set fallback attribute {fallback_key} = {processed_value}"
                    )
                else:
                    span.set_attribute(fallback_key, str(value))
                    logger.info(
                        f"Set fallback attribute {fallback_key} = {value}"
                    )

    def _should_process_as_array(self, attr_constant: str) -> bool:
        """
        Determine if a span attribute constant should be processed as an array.

        Args:
            attr_constant: The span attribute constant (e.g., SpanAttributes.RETRIEVAL.RETRIEVED_CONTEXTS)

        Returns:
            bool: True if the attribute should be processed as an array
        """
        # Define attributes that should be treated as arrays based on semantic conventions
        array_attributes = {
            SpanAttributes.RETRIEVAL.RETRIEVED_CONTEXTS,
            SpanAttributes.GRAPH_NODE.NODES_EXECUTED,
            SpanAttributes.RERANKER.INPUT_CONTEXT_TEXTS,
            SpanAttributes.RERANKER.INPUT_CONTEXT_SCORES,
            SpanAttributes.RERANKER.INPUT_RANKS,
            SpanAttributes.RERANKER.OUTPUT_RANKS,
            SpanAttributes.RERANKER.OUTPUT_CONTEXT_TEXTS,
            SpanAttributes.RERANKER.OUTPUT_CONTEXT_SCORES,
        }
        return attr_constant in array_attributes

    def _should_process_as_array_by_name(self, attr_name: str) -> bool:
        """
        Determine if an attribute should be processed as an array based on its name.
        This is used for fallback cases where we don't have the constant.

        Args:
            attr_name: The attribute name (e.g., "retrieved_contexts")

        Returns:
            bool: True if the attribute should be processed as an array
        """
        # Attribute names that suggest array content
        array_attribute_names = {
            "retrieved_contexts",
            "nodes_executed",
            "input_context_texts",
            "input_context_scores",
            "input_ranks",
            "output_ranks",
            "output_context_texts",
            "output_context_scores",
        }
        return attr_name.lower() in array_attribute_names

    def _process_array_attribute(self, value: any) -> List[str]:
        """
        Process a value that should be treated as an array attribute.

        Args:
            value: The raw value from the dataset, could be a string, list, or other type

        Returns:
            List[str]: A list of string values
        """
        if value is None:
            return []

        # If it's already a list, return it (converting items to strings)
        if isinstance(value, list):
            return [str(item) for item in value]

        # If it's a string, try to parse it as JSON first, then fall back to comma-separated
        if isinstance(value, str):
            value = value.strip()
            if not value:
                return []

            # Try to parse as JSON array first
            try:
                parsed = json.loads(value)
                if isinstance(parsed, list):
                    return [str(item) for item in parsed]
            except (json.JSONDecodeError, ValueError):
                pass

            # Fall back to comma-separated parsing
            # Split by comma and clean up whitespace
            items = [item.strip() for item in value.split(",")]
            # Remove empty strings
            items = [item for item in items if item]
            return items

        # For any other type, convert to string and treat as single item
        return [str(value)]

    def _group_dataset_spec_by_span_type(
        self, dataset_spec: Dict[str, str], row
    ) -> Dict[str, Dict[str, any]]:
        """Group dataset_spec entries by span type (record_root, retrieval, generation, etc.)"""
        span_data = {}

        for spec_key, column_name in dataset_spec.items():
            if column_name not in row:
                continue

            value = row[column_name]
            parts = spec_key.lower().split(".")

            if len(parts) >= 2:
                span_type = f"{parts[0]}"  # e.g., 'record_root'
                attribute = ".".join(
                    parts[1:]
                )  # e.g., 'input', 'output', 'query_text'
            else:
                logger.warning(
                    f"Legacy key: {spec_key} - assuming it's record_root "
                )
                span_type = "record_root"
                attribute = parts[0]

            if span_type not in span_data:
                span_data[span_type] = {}

            span_data[span_type][attribute] = value

        return span_data

    def _get_span_type_enum(self, span_type: str):
        """Map span type string to SpanType enum dynamically"""
        # Convert span_type to uppercase to match enum naming convention
        enum_name = span_type.upper()

        # Try to get the enum value dynamically
        return getattr(SpanAttributes.SpanType, enum_name, None)

    def _get_span_attributes_class(self, span_type: str):
        """Get the appropriate SpanAttributes class for a span type dynamically"""
        # Convert span_type to uppercase to match class naming convention
        class_name = span_type.upper()

        # Try to get the attributes class dynamically
        return getattr(SpanAttributes, class_name, None)

    def _get_current_time_in_ms(self) -> int:
        return int(round(time.time() * 1000))

    def get_status(self) -> RunStatus:
        run_metadata_df = self.run_dao.get_run(
            run_name=self.run_name,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
        )

        run = Run.from_metadata_df(
            run_metadata_df,
            {
                "app": self,
                "main_method_name": self.main_method_name,
                "run_dao": self.run_dao,
                "tru_session": self.tru_session,
            },
        )

        if run.run_status == "CANCELLED":
            return RunStatus.CANCELLED

        if not self._is_invocation_started(run):
            logger.info("Run is created, no invocation nor computation yet.")
            return RunStatus.CREATED
        elif self._metrics_computation_started(run):
            logger.info(
                "Run is created, invocation done, and some or all metrics computed."
            )

            return self._compute_overall_computations_status(run)

        else:
            logger.info("Run is created, invocation started.")

            return self._compute_latest_invocation_status(run)

    def _should_skip_computation(self, metric_name: str, run: Run) -> bool:
        metrics_metadata = (
            run.describe().get("run_metadata", {}).get("metrics", {})
        )

        if metrics_metadata is None:
            return False

        statuses = []  # will store statuses for all matching metric entries
        for metric_metadata in metrics_metadata.values():
            if metric_metadata.get("name", "") == metric_name:
                statuses.append(
                    metric_metadata.get("completion_status", {}).get("status")
                )

        # If no matching metric entries found, don't skip.
        if not statuses:
            return False

        if any(s == Run.CompletionStatusStatus.COMPLETED for s in statuses):
            logger.info(
                f"Metric {metric_name} already computed successfully (one entry COMPLETED); skipping computation."
            )
            return True

        # If any metric is in progress (i.e. started), we skip because it's still computing.
        if any(s == Run.CompletionStatusStatus.STARTED for s in statuses):
            logger.info(
                f"Metric {metric_name} is in progress (at least one entry not complete); skipping computation."
            )
            return True

        # If all matching metrics are FAILED, then allow re-computation.
        if all(s == Run.CompletionStatusStatus.FAILED for s in statuses):
            logger.info(
                f"All metric entries for {metric_name} have FAILED; allowing re-computation."
            )
            return False

        logger.warning(
            "Unknown state for metric computation; skipping computation."
        )
        return True

    def compute_metrics(self, metrics: List[Union[str, MetricConfig]]) -> str:
        """
        Compute metrics for the run.

        Args:
            metrics: List of metric identifiers (strings) for server-side metrics,
                    or MetricConfig objects for client-side metrics

        Returns:
            Status message indicating computation progress
        """
        if not metrics:
            raise ValueError(
                "No metrics provided. Please provide at least one metric to compute."
            )

        run_status = self.get_status()

        logger.info(f"Current run status: {run_status}")

        if not self._can_start_new_metric_computation(run_status):
            return f"""Cannot start a new metric computation when in run status: {run_status}. Valid statuses are: {RunStatus.INVOCATION_COMPLETED}, {RunStatus.INVOCATION_PARTIALLY_COMPLETED},
        {RunStatus.COMPUTATION_IN_PROGRESS}, {RunStatus.COMPLETED}, {RunStatus.PARTIALLY_COMPLETED}, {RunStatus.FAILED}."""

        computed_metrics = []
        for metric in metrics:
            if self._should_skip_computation(metric, self):
                computed_metrics.append(metric)

        if computed_metrics:
            return (
                f"Cannot compute metrics because the following metric(s) are already computed or in progress: "
                f"{', '.join(computed_metrics)}. If you want to recompute, please cancel the run and start a new one."
            )

        # Separate client-side and server-side metrics
        client_metric_configs = []
        server_metric_names = []

        for metric in metrics:
            if isinstance(metric, str):
                # String metrics are server-side
                server_metric_names.append(metric)
            else:
                # MetricConfig objects are client-side
                if (
                    hasattr(metric, "computation_type")
                    and metric.computation_type == "client"
                ):
                    client_metric_configs.append(metric)
                else:
                    # Default to client-side for MetricConfig objects
                    client_metric_configs.append(metric)

        logger.info(
            f"Client-side metrics to compute: {[m.name for m in client_metric_configs]}"
        )
        logger.info(f"Server-side metrics to compute: {server_metric_names}")

        try:
            # Handle client-side metrics
            if client_metric_configs:
                try:
                    self._compute_client_side_metrics_from_configs(
                        client_metric_configs
                    )
                    logger.info(
                        f"Successfully computed {len(client_metric_configs)} client-side metrics"
                    )
                except Exception as e:
                    logger.error(f"Error computing client-side metrics: {e}")
                    raise

            if server_metric_names:
                self.run_dao.call_compute_metrics_query(
                    metrics=server_metric_names,
                    object_name=self.object_name,
                    object_type=self.object_type,
                    object_version=self.object_version,
                    run_name=self.run_name,
                )
                logger.info(
                    f"Started server-side computation for {len(server_metric_names)} metrics"
                )

            logger.info("Metrics computation job started")
        finally:
            self.tru_session.force_flush()

            logger.debug(
                "Flushed OTel eval spans to event table to ensure all spans are ingested before main process exits"
            )

        return "Metrics computation in progress."

    def _compute_client_side_metrics_from_configs(
        self,
        metric_configs: List[Union[MetricConfig, "metric_module.Metric"]],
    ) -> None:
        """Compute client-side custom metrics from Metric or MetricConfig objects."""
        try:
            from trulens.feedback.computer import compute_feedback_by_span_group
        except ImportError:
            logger.error(
                "trulens.feedback package is not installed. Please install it to use feedback computation functionality."
            )
            raise

        # Force flush to ensure spans are uploaded to Snowflake before querying
        self.tru_session.force_flush()
        logger.debug(
            "Flushed OTel spans to event table before retrieving them for client-side metric computation"
        )

        events = self._get_events_for_client_metrics()

        if events.empty:
            logger.warning(
                f"No events found for app {self.app.app_name} version {self.app.app_version} run {self.run_name}"
            )
            return

        # Compute each client-side metric

        for metric_config in metric_configs:
            try:
                logger.info(
                    f"Computing client-side metric: {metric_config.name}"
                )

                # Handle both Metric objects (new API) and MetricConfig (deprecated)
                if isinstance(metric_config, metric_module.Metric):
                    # Metric objects are already feedback definitions
                    feedback = metric_config
                elif hasattr(metric_config, "create_feedback_definition"):
                    # MetricConfig (deprecated) needs conversion
                    feedback = metric_config.create_feedback_definition()
                else:
                    raise TypeError(
                        f"Expected Metric or MetricConfig, got {type(metric_config)}"
                    )

                compute_feedback_by_span_group(
                    events=events,
                    feedback=feedback,
                    raise_error_on_no_feedbacks_computed=False,
                    selectors=feedback.selectors,
                )
                logger.info(
                    f"Successfully computed client-side metric: {metric_config.name}"
                )
            except Exception as e:
                logger.error(
                    f"Error computing client-side metric {metric_config.name}: {e}"
                )
                raise

    def _get_events_for_client_metrics(self) -> pd.DataFrame:
        """Get events for client-side metric computation using the appropriate method."""
        try:
            from trulens.connectors.snowflake import SnowflakeConnector

            if (
                isinstance(self.tru_session.connector, SnowflakeConnector)
                and self.tru_session.connector.use_account_event_table
            ):
                events_df = self.tru_session.connector.db.get_events(
                    app_name=self.app.app_name,
                    app_version=self.app.app_version,
                    run_name=self.run_name,
                )

                if not events_df.empty:
                    for json_col in [
                        "TRACE",
                        "RESOURCE_ATTRIBUTES",
                        "RECORD",
                        "RECORD_ATTRIBUTES",
                    ]:
                        if json_col in events_df.columns:
                            events_df[json_col] = events_df[json_col].apply(
                                json.loads
                            )

                    # Rename columns to match expected format (lowercase) in downstream feedback computation
                    # TODO: remove this once we have a more robust/general way to handle the column names
                    column_mapping = {
                        "TRACE": "trace",
                        "RESOURCE_ATTRIBUTES": "resource_attributes",
                        "RECORD": "record",
                        "RECORD_ATTRIBUTES": "record_attributes",
                    }

                    events_df = events_df.rename(columns=column_mapping)

                    for idx, row in events_df.iterrows():
                        trace = events_df.at[idx, "trace"]
                        record = events_df.at[idx, "record"]
                        record_attributes = (
                            events_df.at[idx, "record_attributes"]
                            if "record_attributes" in events_df.columns
                            else {}
                        )

                        # Parse JSON fields using helper function
                        try:
                            trace = _parse_json_field(
                                trace, "trace", idx, critical=True
                            )
                            events_df.at[idx, "trace"] = trace
                        except json.JSONDecodeError:
                            continue

                        try:
                            record = _parse_json_field(
                                record, "record", idx, critical=True
                            )
                            events_df.at[idx, "record"] = record
                        except json.JSONDecodeError:
                            continue

                        # record_attributes parsing is not critical for parent_id assignment
                        record_attributes = _parse_json_field(
                            record_attributes,
                            "record_attributes",
                            idx,
                            critical=False,
                        )
                        events_df.at[idx, "record_attributes"] = (
                            record_attributes
                        )

                        # Now modify the dictionary
                        if isinstance(trace, dict) and isinstance(record, dict):
                            if "parent_span_id" in record:
                                trace["parent_id"] = record["parent_span_id"]
                            else:
                                trace["parent_id"] = None

                            # Set the modified dict back into the DataFrame
                            events_df.at[idx, "trace"] = trace

                if not events_df.empty and self.run_name:
                    filtered_events = []
                    for _, row in events_df.iterrows():
                        try:
                            record_attributes = row.get("record_attributes", {})

                            # Parse record_attributes using helper function
                            original_record_attributes = record_attributes
                            record_attributes = _parse_json_field(
                                record_attributes,
                                "record_attributes",
                                critical=False,
                            )
                            # If parsing failed and we still have a string, skip this record
                            if (
                                isinstance(record_attributes, str)
                                and record_attributes
                                == original_record_attributes
                            ):
                                continue

                            event_run_name = record_attributes.get(
                                SpanAttributes.RUN_NAME
                            )
                            if event_run_name == self.run_name:
                                filtered_events.append(row)

                        except Exception as e:
                            logger.debug(
                                f"Skipping event due to parsing error: {e}"
                            )
                            continue

                    if filtered_events:
                        events_df = pd.DataFrame(filtered_events)
                        logger.info(
                            f"Filtered {len(filtered_events)} events for run {self.run_name}"
                        )
                    else:
                        logger.warning(
                            f"No events found for run {self.run_name} after filtering"
                        )
                        events_df = pd.DataFrame()

                return events_df
            else:
                return self.tru_session.connector.get_events(
                    app_name=self.app.app_name,
                    app_version=self.app.app_version,
                    run_name=self.run_name,
                )
        except ImportError:
            raise ValueError(
                "Snowflake connector is not installed. Please install it to use feedback computation functionality."
            )

    def get_records(
        self,
        record_ids: Optional[List[str]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> pd.DataFrame:
        """
        A wrapper API around get_records_and_feedback to retrieve and display overview of records from event table of the run.
        It aggregates summary information of records into a single DataFrame.

        Args:
            record_ids: Optional list of record IDs to filter by. Defaults to None.
            offset: Record row offset.
            limit: Limit on the number of records to return.

        Returns:
            A DataFrame with the overview of records.
        """
        record_details_df, metrics_columns = (
            self.tru_session.get_records_and_feedback(
                app_name=self.object_name,
                app_version=self.object_version,
                run_name=self.run_name,
                record_ids=record_ids,
                offset=offset,
                limit=limit,
            )
        )

        record_overview_col_names = [
            "record_id",
            "input",
            "output",
            "latency",
        ] + metrics_columns
        return record_details_df[record_overview_col_names]

    def get_record_details(
        self,
        record_ids: Optional[List[str]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> pd.DataFrame:
        """
        A wrapper API around get_records_and_feedback to retrieve records from event table of the run.

        Args:
            record_ids: Optional list of record IDs to filter by. Defaults to None.
            offset: Record row offset.
            limit: Limit on the number of records to return.

        Returns:
            A DataFrame with the details of records.
        """
        record_details_df, _ = self.tru_session.get_records_and_feedback(
            app_name=self.object_name,
            app_version=self.object_version,
            run_name=self.run_name,
            record_ids=record_ids,
            offset=offset,
            limit=limit,
        )

        return record_details_df

    def _is_cancelled(self) -> bool:
        return self.get_status() == RunStatus.CANCELLED

    def cancel(self):
        if self._is_cancelled():
            logger.warning(f"Run {self.run_name} is already cancelled.")
            return

        update_fields = {"run_status": "CANCELLED"}

        self.run_dao.upsert_run_metadata_fields(
            run_name=self.run_name,
            object_name=self.object_name,
            object_type=self.object_type,
            object_version=self.object_version,
            **update_fields,
        )

        logger.info(f"Run {self.run_name} cancelled.")

    def update(
        self, description: Optional[str] = None, label: Optional[str] = None
    ):
        """
        Only description and label are allowed to be updated at the moment.
        """
        update_fields = {}
        if description is not None:
            logger.info(f"Updating run description to {description}")
            update_fields["description"] = description
        if label is not None:
            logger.info(f"Updating run label to {label}")
            update_fields["labels"] = [label]

        if update_fields:
            self.run_dao.upsert_run_metadata_fields(
                run_name=self.run_name,
                object_name=self.object_name,
                object_type=self.object_type,
                object_version=self.object_version,
                **update_fields,
            )

    @classmethod
    def from_metadata_df(
        cls, metadata_df: pd.DataFrame, extra: Dict[str, Any]
    ) -> Run:
        """
        Create a Run instance from a metadata DataFrame returned by the DAO,
        and enrich it with additional fields (which are not persisted on the server).

        Args:
            metadata_df: A pandas DataFrame containing run metadata.
                We assume the first row contains a JSON string in its first cell.
            extra: A dictionary of extra fields to add, such as:
                {
                    "app": <app instance>,
                    "main_method_name": <method name>,
                    "run_dao": <dao instance>,
                    "object_name": <object name>,
                    "object_type": <object type>
                }

        Returns:
            A validated Run instance.
        """
        if metadata_df.empty:
            raise ValueError("No run metadata found.")

        # Assume the first cell of the first row contains the JSON string.

        metadata_str = metadata_df.iloc[0].values[0]
        try:
            metadata = json.loads(metadata_str)
        except json.JSONDecodeError:
            raise ValueError(
                "The first cell of the first row does not contain a valid JSON string."
            )

        metadata.update(extra)

        return cls.model_validate(metadata)



from __future__ import annotations

import datetime
import logging
import threading
import time
import traceback
from typing import TYPE_CHECKING, Dict, List, Optional
import weakref

import pandas as pd
from trulens.core.otel.utils import is_otel_tracing_enabled
from trulens.core.session import TruSession
from trulens.otel.semconv.trace import SpanAttributes

if TYPE_CHECKING:
    from trulens.core.app import App

logger = logging.getLogger(__name__)

# When computing feedbacks, we only consider events that ended after a certain
# time so that we don't have to routinely scan all the events. Unfortunately,
# the event table doesn't have any timestamp for when a row was added so we use
# the "TIMESTAMP" column which is when the span/event ended.
# This is still problematic because a span can end and then take a while to be
# ingested into the event table. To get around this, we subtract a time delta
# from the last processed time to allow for some leeway.
_PROCESSED_TIME_DELTA = datetime.timedelta(hours=1)


class Evaluator:
    def __init__(self, app: App):
        self._app_ref = weakref.ref(app)
        self._app_name = app.app_name
        self._app_version = app.app_version
        self._thread = None
        self._stop_event = threading.Event()
        self._compute_feedbacks_lock = threading.Lock()
        self._record_id_to_event_count = pd.Series(dtype=int)
        self._processed_time = None

    def _events_under_record_root(self, events: pd.DataFrame) -> pd.DataFrame:
        """
        Get events that are under the record root.

        Args:
            events: A pandas DataFrame of events.

        Returns:
            The events that are under the record root.
        """
        # Construct a tree of events.
        record_roots = []
        span_id_to_children = {
            event["trace"]["span_id"]: [] for _, event in events.iterrows()
        }
        for i, event in events.iterrows():
            parent_id = event["trace"]["parent_id"]
            if parent_id and parent_id in span_id_to_children:
                span_id_to_children[parent_id].append(event)
            if (
                event["record_attributes"].get(SpanAttributes.SPAN_TYPE)
                == SpanAttributes.SpanType.RECORD_ROOT
            ):
                record_roots.append(event)
        # Get events under the record root.
        if len(record_roots) != 1:
            return pd.DataFrame()
        ret = []
        q = [record_roots[0]]
        while q:
            curr_event = q.pop(0)
            ret.append(curr_event)
            q.extend([
                child_event
                for child_event in span_id_to_children[
                    curr_event["trace"]["span_id"]
                ]
            ])
        return pd.DataFrame(ret)

    def _get_record_id_to_unprocessed_events(
        self,
        record_ids: Optional[List[str]],
        start_time: Optional[datetime.datetime],
    ) -> Dict[str, pd.DataFrame]:
        """
        Get events for the app that weren't yet used for feedback computation.

        Args:
            record_ids:
                Optional list of record IDs to filter events by. If None, all
                unprocessed events will be returned.

        Returns:
            A dict from record id to a pandas DataFrame of all events from that
            record. Only records that aren't fully processed will be included.
        """
        # When specific record_ids are provided, don't apply start_time filter.
        # We want ALL events for those specific records regardless of when they
        # were created. The start_time filter is only useful when scanning for
        # new records (record_ids=None) to avoid re-processing old events.
        effective_start_time = None if record_ids is not None else start_time
        events = self._app_ref().connector.get_events(
            app_name=self._app_ref().app_name,
            app_version=self._app_ref().app_version,
            record_ids=record_ids,
            start_time=effective_start_time,
        )
        if events is None or len(events) == 0:
            return {}
        record_ids = events["record_attributes"].apply(
            lambda curr: curr.get(SpanAttributes.RECORD_ID)
        )
        record_id_to_events = events.groupby(record_ids)
        record_id_to_events_under_record_root = {
            k: self._events_under_record_root(events.loc[v])
            for k, v in record_id_to_events.groups.items()
        }
        ret = {}
        for (
            record_id,
            events_under_record_root,
        ) in record_id_to_events_under_record_root.items():
            count = len(events_under_record_root)
            if (
                record_id not in self._record_id_to_event_count
                or count > self._record_id_to_event_count[record_id]
            ):
                ret[record_id] = events_under_record_root
        return ret

    def _compute_feedbacks(
        self,
        record_ids: Optional[List[str]] = None,
        in_evaluator_thread: bool = True,
        lock: Optional[threading.Lock] = None,
    ) -> None:
        new_processed_time = datetime.datetime.now()
        if lock is None:
            lock = self._compute_feedbacks_lock
        with lock:
            if self._processed_time is None:
                logger.info("Processing all events.")
            else:
                logger.info(
                    f"Processing all events from {self._processed_time}"
                )
            record_id_to_events = self._get_record_id_to_unprocessed_events(
                record_ids,
                self._processed_time,
            )
            for record_id, events in record_id_to_events.items():
                try:
                    self._app_ref().compute_feedbacks(
                        raise_error_on_no_feedbacks_computed=False,
                        events=events,
                    )
                except Exception as e:
                    logger.warning(
                        f"Error computing feedbacks in evaluator thread (record_id={record_id}): {e}\n{traceback.format_exc()}"
                    )
                finally:
                    self._record_id_to_event_count[record_id] = len(events)
                    TruSession().force_flush()
                if in_evaluator_thread and self._stop_event.is_set():
                    break
        if not record_ids:
            self._processed_time = new_processed_time - _PROCESSED_TIME_DELTA

    def _run_evaluator(self) -> None:
        """Background thread that periodically computes feedback for events."""
        try:
            while not self._stop_event.is_set():
                self._compute_feedbacks()
                for _ in range(100):
                    if self._stop_event.is_set():
                        break
                    time.sleep(0.1)
        except Exception as e:
            logger.error(f"Evaluator thread encountered an error: {e}")

    def start_evaluator(self) -> None:
        """Start the evaluator for the app."""
        # Validate.
        if not is_otel_tracing_enabled():
            raise ValueError(
                "This method is only supported for OTEL Tracing. Please enable OTEL tracing in the environment!"
            )
        if self._thread is not None:
            raise RuntimeError(
                "Evaluator thread already started. Please stop it before starting a new one."
            )
        # Create and start the evaluator thread.
        self._stop_event.clear()
        self._thread = threading.Thread(
            target=self._run_evaluator,
            daemon=True,
            name=f"evaluator_thread(app_name={self._app_name}, app_version={self._app_version})",
        )
        self._thread.start()

    def stop_evaluator(self) -> None:
        """Stop the evaluator for the app.

        This is only supported for OTEL Tracing.
        """
        if self._thread is None:
            logger.warning("No evaluator thread is running.")
            return
        # Signal the thread to stop.
        self._stop_event.set()
        # If called from within the evaluator thread, skip join to avoid deadlock.
        if threading.current_thread() is self._thread:
            logger.info(
                f"Stop requested from evaluator thread itself; skipping join for (app_name={self._app_name}, app_version={self._app_version})."
            )
        else:
            # Give the thread a reasonable time to exit gracefully.
            self._thread.join(timeout=300)
            # If thread is still alive after timeout, log a warning.
            if self._thread.is_alive():
                logger.warning(
                    f"Evaluator thread (app_name={self._app_name}, app_version={self._app_version}) did not terminate gracefully within timeout."
                )
            else:
                logger.info(
                    f"Stopped evaluator thread (app_name={self._app_name}, app_version={self._app_version})."
                )
        # Reset for potential future restart.
        self._thread = None
        self._stop_event.clear()

    def compute_now(
        self,
        record_ids: Optional[List[str]],
        lock: Optional[threading.Lock] = None,
    ) -> None:
        """Trigger immediate computation.

        Args:
            record_ids:
                Optional list of record ids to compute feedbacks for. If None,
                computes feedbacks for all unprocessed records.
            lock:
                Optional lock to use for the computation. If None, will use the
                default lock.
        """
        self._compute_feedbacks(
            record_ids, in_evaluator_thread=False, lock=lock
        )

    def __del__(self):
        try:
            if self._thread is not None:
                logger.info(
                    f"Stopping evaluator thread during garbage collection (app_name={self._app_name}, app_version={self._app_version})."
                )
                self.stop_evaluator()
        except Exception:
            # During interpreter shutdown, some modules might be already
            # unloaded so we can't rely on the logger or other modules being
            # available
            pass



"""
# Virtual Apps

This module facilitates the ingestion and evaluation of application logs that
were generated outside of TruLens. It allows for the creation of a virtual
representation of your application, enabling the evaluation of logged data
within the TruLens framework.

To begin, construct a virtual application representation. This can be
achieved through a simple dictionary or by utilizing the `VirtualApp` class,
which allows for a more structured approach to storing application
information relevant for feedback evaluation.

Example: Constructing a Virtual Application
    ```python
    virtual_app = {
        'llm': {'modelname': 'some llm component model name'},
        'template': 'information about the template used in the app',
        'debug': 'optional fields for additional debugging information'
    }
    # Converting the dictionary to a VirtualApp instance
    from trulens.core import Select
    from trulens.apps.virtual import VirtualApp

    virtual_app = VirtualApp(virtual_app)
    virtual_app[Select.RecordCalls.llm.maxtokens] = 1024
    ```

Incorporate components into the virtual app for evaluation by utilizing the
`Select` class. This approach allows for the reuse of setup configurations
when defining feedback functions.

Example: Incorporating Components into the Virtual App
    ```python
    # Setting up a virtual app with a retriever component
    from trulens.core import Select
    retriever_component = Select.RecordCalls.retriever
    virtual_app[retriever_component] = 'this is the retriever component'
    ```

With your virtual app configured, it's ready to store logged data.
`VirtualRecord` offers a structured way to build records from your data for
ingestion into TruLens, distinguishing itself from direct `Record` creation
by specifying calls through selectors.

Below is an example of adding records for a context retrieval component,
emphasizing that only the data intended for tracking or evaluation needs to
be provided.

Example: Adding Records for a Context Retrieval Component
    ```python
    from trulens.apps.virtual import VirtualRecord

    # Selector for the context retrieval component's `get_context` call
    context_call = retriever_component.get_context

    # Creating virtual records
    rec1 = VirtualRecord(
        main_input='Where is Germany?',
        main_output='Germany is in Europe',
        calls={
            context_call: {
                'args': ['Where is Germany?'],
                'rets': ['Germany is a country located in Europe.']
            }
        }
    )
    rec2 = VirtualRecord(
        main_input='Where is Germany?',
        main_output='Poland is in Europe',
        calls={
            context_call: {
                'args': ['Where is Germany?'],
                'rets': ['Poland is a country located in Europe.']
            }
        }
    )

    data = [rec1, rec2]
    ```

For existing datasets, such as a dataframe of prompts, contexts, and
responses, iterate through the dataframe to create virtual records for each
entry.

Example: Creating Virtual Records from a DataFrame
    ```python
    import pandas as pd

    # Example dataframe
    data = {
        'prompt': ['Where is Germany?', 'What is the capital of France?'],
        'response': ['Germany is in Europe', 'The capital of France is Paris'],
        'context': [
            'Germany is a country located in Europe.',
            'France is a country in Europe and its capital is Paris.'
        ]
    }
    df = pd.DataFrame(data)

    # Ingesting data from the dataframe into virtual records
    data_dict = df.to_dict('records')
    data = []

    for record in data_dict:
        rec = VirtualRecord(
            main_input=record['prompt'],
            main_output=record['response'],
            calls={
                context_call: {
                    'args': [record['prompt']],
                    'rets': [record['context']]
                }
            }
        )
        data.append(rec)
    ```

After constructing the virtual records, feedback functions can be developed
in the same manner as with non-virtual applications, using the newly added
`context_call` selector for reference. The same process can be repeated for
any additional selector you add.

Example: Developing Feedback Functions
    ```python
    from trulens.providers.openai import OpenAI
    from trulens.core.feedback.feedback import Feedback

    # Initializing the feedback provider
    openai = OpenAI()

    # Defining the context for feedback using the virtual `get_context` call
    context = context_call.rets[:]

    # Creating a feedback function for context relevance
    f_context_relevance = Feedback(openai.context_relevance).on_input().on(context)
    ```

These feedback functions are then integrated into `TruVirtual` to construct
the recorder, which can handle most configurations applicable to non-virtual
apps.

Example: Integrating Feedback Functions into TruVirtual
    ```python
    from trulens.apps.virtual import TruVirtual

    # Setting up the virtual recorder
    virtual_recorder = TruVirtual(
        app_name='a virtual app',
        app_version='base',
        app=virtual_app,
        feedbacks=[f_context_relevance]
    )
    ```

To process the records and run any feedback functions associated with the
recorder, use the `add_record` method.

Example: "Logging records and running feedback functions"

    ```python
    # Ingesting records into the virtual recorder
    for record in data:
        virtual_recorder.add_record(record)
    ```

Metadata about your application can also be included in the `VirtualApp` for
evaluation purposes, offering a flexible way to store additional information
about the components of an LLM app.

Example: Storing metadata in a VirtualApp
    ```python
    # Example of storing metadata in a VirtualApp
    virtual_app = {
        'llm': {'modelname': 'some llm component model name'},
        'template': 'information about the template used in the app',
        'debug': 'optional debugging information'
    }

    from trulens.core import Select
    from trulens.apps.virtual import VirtualApp

    virtual_app = VirtualApp(virtual_app)
    virtual_app[Select.RecordCalls.llm.maxtokens] = 1024
    ```

This approach is particularly beneficial for evaluating the components of an LLM app.

Example: Evaluating components of an LLM application
    ```python
    # Adding a retriever component to the virtual app
    retriever_component = Select.RecordCalls.retriever
    virtual_app[retriever_component] = 'this is the retriever component'
    ```
"""

from concurrent import futures
import datetime
import logging
from pprint import PrettyPrinter
from typing import Any, ClassVar, Dict, List, Optional, Sequence, Union

from pydantic import Field
from trulens.core import app as core_app
from trulens.core import instruments as core_instruments
from trulens.core.schema import base as base_schema
from trulens.core.schema import feedback as feedback_schema
from trulens.core.schema import record as record_schema
from trulens.core.schema import select as select_schema
from trulens.core.utils import pyschema as pyschema_utils
from trulens.core.utils import serial as serial_utils

logger = logging.getLogger(__name__)

pp = PrettyPrinter()


class VirtualApp(dict):
    """A dictionary meant to represent the components of a virtual app.

    `TruVirtual` will refer to this class as the wrapped app. All calls will be
    under `VirtualApp.root`
    """

    @classmethod
    def select_context(cls):
        """Select the context of the virtual app. This is fixed to return the default path."""
        return select_schema.Select.RecordCalls.retriever.get_context.rets[:]

    def __setitem__(
        self, __name: Union[str, serial_utils.Lens], __value: Any
    ) -> None:
        """
        Allow setitem to work on Lenses instead of just strings. Uses `Lens.set`
        if a lens is given.
        """

        if isinstance(__name, str):
            return super().__setitem__(__name, __value)

        # Chop off __app__ or __record__ prefix if there.
        __name = select_schema.Select.dequalify(__name)

        # Chop off "app" prefix if there.
        if (
            isinstance(__name.path[0], serial_utils.GetItemOrAttribute)
            and __name.path[0].get_item_or_attribute() == "app"
        ):
            __name = serial_utils.Lens(path=__name.path[1:])

        # Does not mutate so need to use dict.update .
        temp = __name.set(self, __value)
        self.update(temp)

    def root(self):
        """All virtual calls will have this on top of the stack as if their app
        was called using this as the main/root method."""

        pass


virtual_module = pyschema_utils.Module(
    package_name="trulens", module_name="trulens.apps.virtual"
)
"""Module to represent the module of virtual apps.

Virtual apps will record this as their module.
"""

virtual_class = pyschema_utils.Class(module=virtual_module, name="VirtualApp")
"""Class to represent the class of virtual apps.

Virtual apps will record this as their class.
"""

virtual_object = pyschema_utils.Obj(cls=virtual_class, id=0)
"""Object to represent instances of virtual apps.

Virtual apps will record this as their instance.
"""

virtual_method_root = pyschema_utils.Method(
    cls=virtual_class, obj=virtual_object, name="root"
)
"""Method call to represent the root call of virtual apps.

Virtual apps will record this as their root call.
"""

virtual_method_call = pyschema_utils.Method(
    cls=virtual_class, obj=virtual_object, name="method_name_not_set"
)
"""Method call to represent virtual app calls that do not provide this
information.

Method name will be replaced by the last attribute in the selector provided by user.
"""


class VirtualRecord(record_schema.Record):
    """Virtual records for virtual apps.

    Many arguments are filled in by default values if not provided. See
    [Record][trulens.core.schema.record.Record] for all arguments. Listing here is
    only for those which are required for this method or filled with default values.

    Args:
        calls: A dictionary of calls to be recorded. The keys are selectors
            and the values are dictionaries with the keys listed in the next
            section.

        cost: Defaults to zero cost.

        perf: Defaults to time spanning the processing of this virtual
            record. Note that individual calls also include perf. Time span
            is extended to make sure it is not of duration zero.

    Call values are dictionaries containing arguments to
    [RecordAppCall][trulens.core.schema.record.RecordAppCall] constructor. Values
    can also be lists of the same. This happens in non-virtual apps when the
    same method is recorded making multiple calls in a single app
    invocation. The following defaults are used if not provided.

    | PARAMETER | TYPE |DEFAULT |
    | --- | ---| --- |
    | `stack` | [List][typing.List][[RecordAppCallMethod][trulens.core.schema.record.RecordAppCallMethod]] | Two frames: a root call followed by a call by [virtual_object][trulens.apps.virtual.virtual_object], method name derived from the last element of the selector of this call. |
    | `args` | [JSON][trulens.core.utils.serial.JSON] | `[]` |
    | `rets` | [JSON][trulens.core.utils.serial.JSON] | `[]` |
    | `perf` | [Perf][trulens.core.schema.base.Perf] | Time spanning the processing of this virtual call. |
    | `pid` | [int][] | `0` |
    | `tid` | [int][] | `0` |
    """

    def __init__(
        self,
        calls: Dict[serial_utils.Lens, Union[Dict, Sequence[Dict]]],
        cost: Optional[base_schema.Cost] = None,
        perf: Optional[base_schema.Perf] = None,
        **kwargs: Any,
    ):
        root_call = record_schema.RecordAppCallMethod(
            path=serial_utils.Lens(), method=virtual_method_root
        )

        record_calls = []

        start_time = datetime.datetime.now()

        for lens, call_or_calls in calls.items():
            if isinstance(call_or_calls, Sequence):
                calls_list = call_or_calls
            else:
                calls_list = [call_or_calls]

            for call in calls_list:
                substart_time = datetime.datetime.now()

                if "stack" not in call:
                    path, method_name = select_schema.Select.path_and_method(
                        select_schema.Select.dequalify(lens)
                    )
                    method = virtual_method_call.replace(name=method_name)

                    call["stack"] = [
                        root_call,
                        record_schema.RecordAppCallMethod(
                            path=path, method=method
                        ),
                    ]

                if "args" not in call:
                    call["args"] = []
                if "rets" not in call:
                    call["rets"] = []
                if "pid" not in call:
                    call["pid"] = 0
                if "tid" not in call:
                    call["tid"] = 0

                subend_time = datetime.datetime.now()

                # NOTE(piotrm for garrett): that the dashboard timeline has problems
                # with calls that span too little time so we add some delays to the
                # recorded perf.
                if (subend_time - substart_time).total_seconds() == 0.0:
                    subend_time += datetime.timedelta(microseconds=1)

                if "perf" not in call:
                    call["perf"] = base_schema.Perf(
                        start_time=substart_time, end_time=subend_time
                    )

                rinfo = record_schema.RecordAppCall(**call)
                record_calls.append(rinfo)

        end_time = datetime.datetime.now()

        # NOTE(piotrm for garrett): that the dashboard timeline has problems
        # with calls that span too little time so we add some delays to the
        # recorded perf.
        if (end_time - start_time).total_seconds() == 0.0:
            end_time += datetime.timedelta(microseconds=1)

        kwargs["cost"] = cost or base_schema.Cost()
        kwargs["perf"] = perf or base_schema.Perf(
            start_time=start_time, end_time=end_time
        )

        if "main_input" not in kwargs:
            kwargs["main_input"] = "No main_input provided."
        if "main_output" not in kwargs:
            kwargs["main_output"] = "No main_output provided."

        # append root call
        record_calls.append(
            record_schema.RecordAppCall(
                stack=[root_call],
                args=[kwargs["main_input"]],
                rets=[kwargs["main_output"]],
                perf=kwargs["perf"],
                cost=kwargs["cost"],
                tid=0,
                pid=0,
            )
        )

        if "app_id" not in kwargs:
            kwargs["app_id"] = (
                "No app_id provided."  # this gets replaced by TruVirtual.add_record .
            )

        super().__init__(calls=record_calls, **kwargs)


class TruVirtual(core_app.App):
    """Recorder for virtual apps.

    Virtual apps are data only in that they cannot be executed but for whom
    previously-computed results can be added using
    [add_record][trulens.apps.virtual.TruVirtual]. The
    [VirtualRecord][trulens.apps.virtual.VirtualRecord] class may be useful
    for creating records for this. Fields used by non-virtual apps can be
    specified here, notably:

    See [App][trulens.core.app.App] and
    [AppDefinition][trulens.core.schema.app.AppDefinition] for constructor
    arguments.

    # The `app` field.

    You can store any information you would like by passing in a dictionary to
    TruVirtual in the `app` field. This may involve an index of components or
    versions, or anything else. You can refer to these values for evaluating
    feedback.

    Usage:
        You can use `VirtualApp` to create the `app` structure or a plain
        dictionary. Using `VirtualApp` lets you use Selectors to define components:

        ```python
        virtual_app = VirtualApp()
        virtual_app[Select.RecordCalls.llm.maxtokens] = 1024
        ```

    Example:
        ```python
        virtual_app = dict(
            llm=dict(
                modelname="some llm component model name"
            ),
            template="information about the template I used in my app",
            debug="all of these fields are completely optional"
        )

        virtual = TruVirtual(
            app_name="my_virtual_app",
            app_version="base",
            app=virtual_app
        )
        ```
    """

    app: VirtualApp = Field(default_factory=VirtualApp)

    root_callable: ClassVar[pyschema_utils.FunctionOrMethod] = (
        virtual_method_root
    )

    root_class: Any = pyschema_utils.Class.of_class(VirtualApp)

    instrument: Optional[core_instruments.Instrument] = None

    selector_check_warning: bool = False
    """Selector checking is disabled for virtual apps."""

    selector_nocheck: bool = True
    """The selector check must be disabled for virtual apps.

    This is because methods that could be called are not known in advance of
    creating virtual records.
    """

    def __init__(
        self,
        app: Optional[Union[VirtualApp, serial_utils.JSON]] = None,
        **kwargs: Any,
    ):
        """Virtual app for logging existing app results."""

        if app is None:
            app = VirtualApp()
        else:
            if isinstance(app, dict):
                app = VirtualApp(app)
            else:
                raise ValueError(
                    "Unknown type for `app`. "
                    "Either dict or `trulens.apps.virtual.VirtualApp` expected."
                )

        if (
            kwargs.get("selector_nocheck") is False
            or kwargs.get("selector_check_warning") is True
        ):
            raise ValueError(
                "Selector prechecking does not work with virtual apps. "
                "The settings `selector_nocheck=True` and `selector_check_warning=False` are required."
            )

        super().__init__(app=app, **kwargs)

    def add_record(
        self,
        record: record_schema.Record,
        feedback_mode: Optional[feedback_schema.FeedbackMode] = None,
    ) -> record_schema.Record:
        """Add the given record to the database and evaluate any pre-specified
        feedbacks on it.

        The class `VirtualRecord` may be useful for creating
        records for virtual models. If `feedback_mode` is specified, will use
        that mode for this record only.
        """

        if feedback_mode is None:
            feedback_mode = self.feedback_mode

        record.app_id = self.app_id

        # Creates feedback futures.
        record.feedback_and_future_results = self._handle_record(
            record, feedback_mode=feedback_mode
        )
        if record.feedback_and_future_results is not None:
            record.feedback_results = [
                tup[1] for tup in record.feedback_and_future_results
            ]

        # Wait for results if mode is WITH_APP.
        if (
            feedback_mode == feedback_schema.FeedbackMode.WITH_APP
            and record.feedback_results is not None
        ):
            futs = record.feedback_results
            futures.wait(futs)

        return record

    def add_dataframe(
        self,
        df,
        feedback_mode: Optional[feedback_schema.FeedbackMode] = None,
    ) -> List[record_schema.Record]:
        """Add the given dataframe as records to the database and evaluate any pre-specified
        feedbacks on them.

        The class `VirtualRecord` may be useful for creating records for virtual models.

        If `feedback_mode` is specified, will use that mode for these records only.
        """

        data_dict = df.to_dict("records")
        records = []

        for record in data_dict:
            if "contexts" in record:
                retriever = select_schema.Select.RecordCalls.retriever
                if retriever not in self.app:
                    self.app[retriever] = None
                self.app[retriever] = record["contexts"]

                context_rets = (
                    record["contexts"]
                    if isinstance(record["contexts"], list)
                    else [record["contexts"]]
                )

                rec = VirtualRecord(
                    calls={
                        retriever.get_context: dict(
                            args=[record["query"]], rets=context_rets
                        )
                    },
                    main_input=record["query"],
                    main_output=record["response"],
                )
            else:
                rec = VirtualRecord(
                    main_input=record["query"],
                    main_output=record["response"],
                    calls={},
                )
            records.append(rec)

        for record in records:
            self.add_record(record, feedback_mode=feedback_mode)

        return records


TruVirtual.model_rebuild()
